"""The "fred/series" API.

This is by far the most popular FRED API implementation.
Useful for examining historical or projected economic data.
The API comes with builtin methods for filtering data based
on publication date or frequency. See the docs of each
method for more details.

See the official FRED API docs for more info:

    https://fred.stlouisfed.org/docs/api/fred/

"""

import pandas as pd
from requests import HTTPError

from . import _api


class SeriesCategories(_api.API):
    """Get the categories for an economic data series.

    The class variable :attr:`finagg.fred.api.Series.categories` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/categories"

    @classmethod
    def get(
        cls,
        series_id: str,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the categories for an economic data series.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_categories.html

        Args:
            series_id: The ID for a series.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data on categories for the economic data series.

        Examples:
            >>> finagg.fred.api.series.categories.get("CPIAUCNS")  # doctest: +SKIP
               id                                  name  parent_id
            0   9  Consumer Price Indexes (CPI and PCE)      32455

        """
        data = _api.get(
            cls.url,
            series_id=series_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            cache=cache,
            api_key=api_key,
        ).json()
        data = data["categories"]
        return pd.DataFrame(data)


class SeriesObservations(_api.API):
    """Get the observations or data values for an economic data series.

    This is by far the most popular FRED API method. The class variable
    :attr:`finagg.fred.api.Series.observations` is an instance of this
    API implementation and is the most popular interface for calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/observations"

    @classmethod
    def get(
        cls,
        series_id: str,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 100000,
        offset: None | int = 0,
        sort_order: None | str = None,
        observation_start: None | int | str = None,
        observation_end: None | int | str = None,
        units: None | str = "lin",
        frequency: None | str = None,
        aggregation_method: None | str = "avg",
        output_type: None | int = 1,
        vintage_dates: None | str | list[str] = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the observations or data values for an economic data series.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_observations.html

        Args:
            series_id: The ID for a series.
            realtime_start: Start date for fetching results
                according to their publication date. ``0`` indicates since
                the beginning of time.
            realtime_end: End date for fetching results according
                to their publication date. ``-1`` indicates to present day.
            limit: Maximum number of results to return.
            offset: Result start offset.
            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            observation_start: The start date of the observation period.
            observation_end: The end date of the observation period.
            units: Units to return the series values in. Options include:

                - "lin" = levels (no unit transformation)
                - "chg" = change
                - "ch1" = change from a year ago
                - "pch" = percent change
                - "pc1" = percent change from a year ago
                - "pca" = compounded annual rate of change
                - "cch" = continuously compounded rate of change
                - "cca" = continuously compounded annual rate of change
                - "log" = natural log

            frequency: An optional parameter that indicates a lower frequency to
                aggregate values to. Frequency options without period descriptions
                include:

                    - "d" = daily
                    - "w" = weekly
                    - "bw" = biweekly
                    - "m" = monthly
                    - "q" = quarterly
                    - "sa" = semiannual
                    - "a" = annual

                Frequency options with period descriptions include:

                    - "wef" = weekly, ending Friday
                    - "weth" = weekly, ending Thursday
                    - "wetu" = weekly, ending Wednesday
                    - "wem" = weekly, ending Monday
                    - "wesu" = weekly, ending Sunday
                    - "wesa" = weekly, ending Saturday
                    - "bwew" = weekly, ending Wednesday
                    - "bwem" = weekly, Monday

            aggregation_method: A key that indicates the aggregation method used
                for frequency aggregation. Options include:

                    - "avg" = average
                    - "sum" = sum
                    - "eop" = end of period

            output_type: An integer indicating the type of observations to include.
                Options include:

                    - 1 = observations by realtime period
                    - 2 = all observations by vintage dates
                    - 3 = new and revised observations only
                    - 4 = initial release observations only

            vintage_dates: Vintage dates used to download data as it existed on these
                specified dates in history. Vintage dates can be specified instead of
                realtime periods.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing economic data series observations/values according to
            the given parameters.

        Examples:
            >>> finagg.fred.api.series.observations.get(
            ...     "CPIAUCNS",
            ...     realtime_start=0,
            ...     realtime_end=-1,
            ...     output_type=4
            ... ).head(5)  # doctest: +SKIP
              realtime_start realtime_end        date  value series_id
            0     1949-04-22   1953-02-26  1949-03-01  169.5  CPIAUCNS
            1     1949-05-23   1953-02-26  1949-04-01  169.7  CPIAUCNS
            2     1949-06-24   1953-02-26  1949-05-01  169.2  CPIAUCNS
            3     1949-07-22   1953-02-26  1949-06-01  169.6  CPIAUCNS
            4     1949-08-26   1953-02-26  1949-07-01  168.5  CPIAUCNS

        """
        df = _api.maybe_paginate(
            "observations",
            cls.url,
            series_id=series_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            sort_order=sort_order,
            observation_start=observation_start,
            observation_end=observation_end,
            units=units,
            frequency=frequency,
            aggregation_method=aggregation_method,
            output_type=output_type,
            vintage_dates=vintage_dates,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )
        df["series_id"] = series_id
        df["value"] = pd.to_numeric(df["value"], errors="coerce")
        return df

    @classmethod
    def get_original_observations(
        cls,
        series_id: str,
        /,
        *,
        limit: None | int = 100000,
        offset: None | int = 0,
        sort_order: None | str = None,
        observation_start: None | int | str = None,
        observation_end: None | int | str = None,
        units: None | str = "lin",
        frequency: None | str = None,
        aggregation_method: None | str = "avg",
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get only the initial releases/observations for an
        economic data series.

        Similar to :meth:`SeriesObservations.get`, but tries to get initial
        releases/observations only.

        Args:
            series_id: The ID for a series.
            limit: Maximum number of results to return.
            offset: Result start offset.
            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            observation_start: The start date of the observation period.
            observation_end: The end date of the observation period.
            units: Units to return the series values in. Options include:

                - "lin" = levels (no unit transformation)
                - "chg" = change
                - "ch1" = change from a year ago
                - "pch" = percent change
                - "pc1" = percent change from a year ago
                - "pca" = compounded annual rate of change
                - "cch" = continuously compounded rate of change
                - "cca" = continuously compounded annual rate of change
                - "log" = natural log

            frequency: An optional parameter that indicates a lower frequency to
                aggregate values to. Frequency options without period descriptions
                include:

                    - "d" = daily
                    - "w" = weekly
                    - "bw" = biweekly
                    - "m" = monthly
                    - "q" = quarterly
                    - "sa" = semiannual
                    - "a" = annual

                Frequency options with period descriptions include:

                    - "wef" = weekly, ending Friday
                    - "weth" = weekly, ending Thursday
                    - "wetu" = weekly, ending Wednesday
                    - "wem" = weekly, ending Monday
                    - "wesu" = weekly, ending Sunday
                    - "wesa" = weekly, ending Saturday
                    - "bwew" = weekly, ending Wednesday
                    - "bwem" = weekly, Monday

            aggregation_method: A key that indicates the aggregation method used
                for frequency aggregation. Options include:

                    - "avg" = average
                    - "sum" = sum
                    - "eop" = end of period

            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing economic data series observations/values according to
            the given parameters.

        Examples:
            >>> finagg.fred.api.series.observations.get_original_observations(
            ...     "CPIAUCNS",
            ... ).head(5)  # doctest: +SKIP
              realtime_start realtime_end        date  value series_id
            0     1949-04-22   1953-02-26  1949-03-01  169.5  CPIAUCNS
            1     1949-05-23   1953-02-26  1949-04-01  169.7  CPIAUCNS
            2     1949-06-24   1953-02-26  1949-05-01  169.2  CPIAUCNS
            3     1949-07-22   1953-02-26  1949-06-01  169.6  CPIAUCNS
            4     1949-08-26   1953-02-26  1949-07-01  168.5  CPIAUCNS

        """
        try:
            return cls.get(
                series_id,
                realtime_start=0,
                realtime_end=-1,
                limit=limit,
                offset=offset,
                sort_order=sort_order,
                observation_start=observation_start,
                observation_end=observation_end,
                units=units,
                frequency=frequency,
                aggregation_method=aggregation_method,
                output_type=4,
                paginate=paginate,
                cache=cache,
                api_key=api_key,
            )
        except HTTPError:
            return cls.get(
                series_id,
                limit=limit,
                offset=offset,
                sort_order=sort_order,
                observation_start=observation_start,
                observation_end=observation_end,
                units=units,
                frequency=frequency,
                aggregation_method=aggregation_method,
                paginate=paginate,
                cache=cache,
                api_key=api_key,
            )


class SeriesRelease(_api.API):
    """Get the latest release for an economic data seris.

    The class variable :attr:`finagg.fred.api.Series.release` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/release"

    @classmethod
    def get(
        cls,
        series_id: str,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the release for an economic data series.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_release.html

        Args:
            series_id: The ID for a series.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data on a release for an economic data series.

        """
        data = _api.get(
            cls.url,
            series_id=series_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            cache=cache,
            api_key=api_key,
        ).json()
        data = data["releases"]
        return pd.DataFrame(data)


class SeriesSearchRelatedTags(_api.API):
    """Search for series tags related to a series's tags.

    The class variable :attr:`finagg.fred.api.SeriesSearch.related_tags` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/search/related_tags"

    @classmethod
    def get(
        cls,
        series_search_text: None | str | list[str],
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        tag_names: None | str | list[str] = None,
        exclude_tag_names: None | str | list[str] = None,
        tag_group_id: None | str = None,
        tag_search_text: None | str | list[str] = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the related tags for a series search.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_search_tags.html

        Args:
            series_search_text: The words to match against economic data series.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            tag_names: Tag names to only include in the response.
            exclude_tag_names: Tag names that series match none of.
            tag_group_id: A tag group ID to filter tags by type. Options include:

                - "freq" = frequency
                - "gen" = general or concept
                - "geo" = geography
                - "geot" = geography type
                - "rls" = release
                - "seas" = seasonal adjustment
                - "src" = source

            tag_search_text: The words to find matching tags with.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Order results by values of the specified attribute.
                Options include:

                    - "series_count"
                    - "popularity"
                    - "created"
                    - "name"
                    - "group_id"

            sort_order: Sort results in ascending ("asc") or descending ("desc")
                order for the attribute values specified by `order_by`.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing related FRED tags for a series search.
            The dataframe can have results optionally filtered by the FRED
            servers according to the method's args.

        Examples:
            >>> finagg.fred.api.series.search.related_tags.get("price index", tag_names="price", limit=5)  # doctest: +SKIP
                                            name group_id                    notes                 created ...
            0                                nsa     seas  Not Seasonally Adjusted  2012-02-27 10:18:19-06 ...
            1                            indexes      gen                           2012-02-27 10:18:19-06 ...
            2                        price index      gen                           2012-02-27 10:18:19-06 ...
            3                            monthly     freq                           2012-02-27 10:18:19-06 ...
            4  public domain: citation requested       cc                     None  2018-12-17 23:33:13-06 ...

        """
        return _api.maybe_paginate(
            "tags",
            cls.url,
            series_search_text=series_search_text,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            tag_names=tag_names,
            exclude_tag_names=exclude_tag_names,
            tag_group_id=tag_group_id,
            tag_search_text=tag_search_text,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class SeriesSearchTags(_api.API):
    """Get FRED series tags.

    The class variable :attr:`finagg.fred.api.SeriesSearch.tags` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/search/tags"

    @classmethod
    def get(
        cls,
        series_search_text: None | str | list[str],
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        tag_names: None | str | list[str] = None,
        tag_group_id: None | str = None,
        tag_search_text: None | str | list[str] = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the tags for a series search.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_search_tags.html

        Args:
            series_search_text: The words to match against economic data series.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            tag_names: Tag names to only include in the response.
            tag_group_id: A tag group ID to filter tags by type. Options include:

                - "freq" = frequency
                - "gen" = general or concept
                - "geo" = geography
                - "geot" = geography type
                - "rls" = release
                - "seas" = seasonal adjustment
                - "src" = source

            tag_search_text: The words to find matching tags with.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Order results by values of the specified attribute.
                Options include:

                    - "series_count"
                    - "popularity"
                    - "created"
                    - "name"
                    - "group_id"

            sort_order: Sort results in ascending ("asc") or descending ("desc")
                order for the attribute values specified by `order_by`.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing FRED tags for a series search.
            The dataframe can have results optionally filtered by the FRED
            servers according to the method's args.

        Examples:
            >>> finagg.fred.api.series.search.tags.get("price index", limit=5)  # doctest: +SKIP
                                            name group_id                    notes ...
            0                                nsa     seas  Not Seasonally Adjusted ...
            1                              price      gen                          ...
            2                            indexes      gen                          ...
            3                        price index      gen                          ...
            4  public domain: citation requested       cc                     None ...

        """
        return _api.maybe_paginate(
            "tags",
            cls.url,
            series_search_text=series_search_text,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            tag_names=tag_names,
            tag_group_id=tag_group_id,
            tag_search_text=tag_search_text,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class SeriesSearch(_api.API):
    """Get economic data series that match search text.

    The class variable :attr:`finagg.fred.api.Series.search` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    related_tags = SeriesSearchRelatedTags()
    """"series/search/related_tags" FRED API. Get the related tags for a
    series search. The most popular way for accessing the
    :class:`finagg.fred.api.SeriesSearchRelatedTags` API.

    :meta hide-value:
    """

    tags = SeriesSearchTags()
    """"series/search/tags" FRED API. Get the tags for a series search.
    The most popular way for accessing the
    :class:`finagg.fred.api.SeriesSearchTags` API.

    :meta hide-value:
    """

    url = "https://api.stlouisfed.org/fred/series/search"

    @classmethod
    def get(
        cls,
        search_text: str | list[str],
        /,
        *,
        search_type: None | str = "full_text",
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        filter_variable: None | str = None,
        filter_value: None | str = None,
        tag_names: None | str | list[str] = None,
        exclude_tag_names: None | str | list[str] = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get economic data series that match search text.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_search.html

        Args:
            search_text: The words to match against economic data series.
            search_type: Determines the type of search to perform. Options include:

                - "full_text" = search series attributes, units, frequency,
                  and tags by parsing words into stems.
                - "series_id" = performs a substring search on series IDs.

            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Order results by values of the specified attribute.
                Options include:

                    - "search_rank"
                    - "series_id"
                    - "title"
                    - "units"
                    - "frequency"
                    - "seasonal_adjustment"
                    - "realtime_start"
                    - "realtime_end"
                    - "last_updated"
                    - "observation_start"
                    - "observation_end"
                    - "popularity"
                    - "group_popularity"

            sort_order: Sort results in ascending ("asc") or descending ("desc")
                order for the attribute values specified by `order_by`.
            filter_variable: The attribute to filter results by. Options include:

                - "frequency"
                - "units"
                - "seasonal_adjustment"

            filter_value: The value of the `filter_variable` attribute to filter
                results by.
            tag_names: List of tag names that series match all of.
            exclude_tag_names: List of tag names that series match none of.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data on series matching the search.

        Examples:
            >>> finagg.fred.api.series.search.get("price index", limit=5)  # doctest: +SKIP
                        id realtime_start realtime_end                                              title ...
            0     CPIAUCSL     2023-03-16   2023-03-16  Consumer Price Index for All Urban Consumers: ... ...
            1     CPIAUCNS     2023-03-16   2023-03-16  Consumer Price Index for All Urban Consumers: ... ...
            2  CUUS0000SA0     2023-03-16   2023-03-16  Consumer Price Index for All Urban Consumers: ... ...
            3   CSUSHPINSA     2023-03-16   2023-03-16    S&P/Case-Shiller U.S. National Home Price Index ...
            4    CSUSHPISA     2023-03-16   2023-03-16    S&P/Case-Shiller U.S. National Home Price Index ...

        """
        return _api.maybe_paginate(
            "seriess",
            cls.url,
            search_text=search_text,
            search_type=search_type,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            filter_variable=filter_variable,
            filter_value=filter_value,
            tag_names=tag_names,
            exclude_tag_names=exclude_tag_names,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class SeriesTags(_api.API):
    """Get FRED tags for an economic data series.

    The class variable :attr:`finagg.fred.api.Series.tags` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/tags"

    @classmethod
    def get(
        cls,
        series_id: str,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        order_by: None | str = None,
        sort_order: None | str = None,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the FRED tags for a series.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_tags.html

        Args:
            series_id: The ID for a series.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            order_by: Order results by values of the specified attribute.
                Options include:

                    - "series_count"
                    - "popularity"
                    - "created"
                    - "name"
                    - "group_id"

            sort_order: Sort results in ascending ("asc") or descending ("desc")
                order for the attribute values specified by `order_by`.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data on FRED tags for series.

        Examples:
            >>> finagg.fred.api.series.tags.get("CPIAUCNS").head(5)  # doctest: +SKIP
                                            name group_id                     notes ...
            0                                nsa     seas   Not Seasonally Adjusted ...
            1                                usa      geo  United States of America ...
            2  public domain: citation requested       cc                      None ...
            3                             nation     geot                           ...
            4                            monthly     freq                           ...

        """
        data = _api.get(
            cls.url,
            series_id=series_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            order_by=order_by,
            sort_order=sort_order,
            cache=cache,
            api_key=api_key,
        ).json()
        data = data["tags"]
        return pd.DataFrame(data)


class SeriesUpdates(_api.API):
    """Get data on when economic data series where updated on the FRED server.

    The class variable :attr:`finagg.fred.api.Series.updates` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/updates"

    @classmethod
    def get(
        cls,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        filter_value: None | str = None,
        start_time: None | str = None,
        end_time: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get economic data series sorted by when observations
        were updated on the FRED server.

        Results are limited to series updated within the last two
        weeks.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_updates.html

        Args:
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            filter_value: Limit results by geographic type of economic data
                series. Options include:

                    - "macro" = limit results to macroeconomic data series
                    - "regional" = limit results to series for parts of the US
                    - "all" = does not filter results

            start_time: Start time for limiting results for a time range.
                Can filter down to minutes. Expects format "YYYMMDDHhmm".
            end_time: Start time for limiting results for a time range.
                Can filter down to minutes. Expects format "YYYMMDDHhmm".
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing info on recently updated economic
            data series.

        """
        return _api.maybe_paginate(
            "seriess",
            cls.url,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            filter_value=filter_value,
            start_time=start_time,
            end_time=end_time,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class SeriesVintageDates(_api.API):
    """Get FRED series revision dates.

    The class variable :attr:`finagg.fred.api.Series.vintage_dates` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/series/vintage_dates"

    @classmethod
    def get(
        cls,
        series_id: str,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 10000,
        offset: None | int = 0,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the dates in history when a series' data values were revised
        or new data values were released.

        Vintage dates are the release dates for a series excluding release dates
        when the data for the series did not change.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series_vintagedates.html

        Args:
            series_id: The ID for a series.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            sort_order: Sort results in ascending ("asc") or descending ("desc")
                vintage date order.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing dates on vintage release dates for a series.

        """
        return _api.maybe_paginate(
            "seriess",
            cls.url,
            series_id=series_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class Series(_api.API):
    """Get an economic data series.

    The class variable :data:`finagg.fred.api.series` is an instance of this
    API implementation and is the most popular interface for calling this API.

    """

    categories = SeriesCategories()
    """"series/categories" FRED API. Get the categories for an economic
    data series. The most popular way for accessing the
    :class:`finagg.fred.api.SeriesCategories` API.

    :meta hide-value:
    """

    observations = SeriesObservations()
    """"series/observations" FRED API. Get the observations or
    data values for an economic data series. The most popular way
    for accessing the :class:`finagg.fred.api.SeriesObservations` API.

    :meta hide-value:
    """

    release = SeriesRelease()
    """"series/release" FRED API. Get the release for an economic data series.
    The most popular way for accessing the
    :class:`finagg.fred.api.SeriesRelease` API.

    :meta hide-value:
    """

    search = SeriesSearch()
    """"series/search" FRED API. Get economic data series that match search
    text. The most popular way for accessing the
    :class:`finagg.fred.api.SeriesSearch` API.

    :meta hide-value:
    """

    tags = SeriesTags()
    """"series/tags" FRED API. Get FRED tags for a series.
    The most popular way for accessing the :class:`finagg.fred.api.SeriesTags`
    API.

    :meta hide-value:
    """

    updates = SeriesUpdates()
    """"series/updates" FRED API. Get economic data series sorted by
    when observations were updated on the FRED server. The most popular
    way for accessing the :class:`finagg.fred.api.SeriesUpdates` API.

    :meta hide-value:
    """

    url = "https://api.stlouisfed.org/fred/series"

    vintage_dates = SeriesVintageDates()
    """"series/vintage_dates" FRED API. Get the dates in history when a
    a series' data values were revised or new data values were released.
    The most popular way for accessing the
    :class:`finagg.fred.api.SeriesVintageDates` API.

    :meta hide-value:
    """

    @classmethod
    def get(
        cls,
        series_id: str,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get an economic data series.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/series.html

        Args:
            series_id: The ID for a series.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing info on an economic data series.

        Examples:
            >>> finagg.fred.api.series.get("CPIAUCNS", realtime_start=0, realtime_end=-1).head(5)  # doctest: +SKIP
                     id realtime_start realtime_end                                              title ...
            0  CPIAUCNS     1949-03-24   1953-02-26  Consumer Price Index for Urban Wage Earners an... ...
            1  CPIAUCNS     1953-02-27   1962-02-27  Consumer Price Index for Urban Wage Earners an... ...
            2  CPIAUCNS     1962-02-28   1971-02-18  Consumer Price Index for Urban Wage Earners an... ...
            3  CPIAUCNS     1971-02-19   1978-02-26  Consumer Price Index for Urban Wage Earners an... ...
            4  CPIAUCNS     1978-02-27   1988-02-25  Consumer Price Index for All Urban Consumers: ... ...

        """
        data = _api.get(
            cls.url,
            series_id=series_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            cache=cache,
            api_key=api_key,
        ).json()
        data = data["seriess"]
        return pd.DataFrame(data)


popular_series = [
    "CIVPART",  # Labor force participation rate
    "CPIAUCNS",  # Consumer price index
    "CSUSHPINSA",  # S&P/Case-Shiller national home price index
    "DJIA",  # Dow Jones Industrial Average index
    "FEDFUNDS",  # Federal funds interest rate
    "GDP",  # Gross domestic product
    "GDPC1",  # Real gross domestic product
    "GS10",  # 10-Year treasury yield
    "M2",  # Money stock measures (i.e., savings and related balances)
    "MICH",  # University of Michigan: inflation expectation
    "NASDAQ100",  # Nasdaq 100 index
    "NASDAQCOM",  # Nasdaq Composite index
    "PSAVERT",  # Personal savings rate
    "SP500",  # S&P 500 index
    "UMCSENT",  # University of Michigan: consumer sentiment
    "UNRATE",  # Unemployment rate
    "WALCL",  # US assets, total assets (less eliminations from consolidation)
]
"""Economic data series that are relatively popular for economic analysis.
Includes things like gross domestic product, unemployment rate, etc..

:meta hide-value:
"""



"""CLI and tools for aggregating fundamental features."""

import logging
import multiprocessing as mp
import os
from typing import Literal

import click

from . import feat as _feat

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


@click.group(help="Fundamental feature tools.")
def entry_point() -> None:
    ...


@entry_point.command(
    help=(
        "Drop and recreate tables, and install recommended tables into the "
        "SQL database."
    ),
)
@click.option(
    "--refined",
    type=click.Choice(["fundam", "fundam.normalized"]),
    multiple=True,
    help=(
        "Refined tables to install. This requires Yahoo! Finance and "
        "SEC refined tables to be installed beforehand."
    ),
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    default=False,
    help="Whether to install all defined tables (including all refined tables).",
)
@click.option(
    "--processes",
    "-n",
    type=int,
    default=mp.cpu_count() - 1,
    help=(
        "Number of backgruond processes to run in parallel when installing data. Note,"
        " not all tables support installations with multiprocessing."
    ),
)
@click.option(
    "--recreate-tables",
    "-r",
    is_flag=True,
    default=False,
    help=(
        "Whether to reset the tables associated with the install options by "
        "dropping and recreating them."
    ),
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Sets the log level to DEBUG to show installation errors for each series.",
)
def install(
    refined: list[Literal["fundam", "fundam.normalized"]] = [],
    all_: bool = False,
    processes: int = mp.cpu_count() - 1,
    recreate_tables: bool = False,
    verbose: bool = False,
) -> int:
    if verbose:
        logging.getLogger(__package__).setLevel(logging.DEBUG)

    if "SEC_API_USER_AGENT" not in os.environ:
        logger.warning(
            "No SEC API user agent found in the environment. "
            "Skipping finagg.fundam installation."
        )
        return 0

    total_rows = 0
    all_refined = set()
    if all_:
        all_refined = {"fundam", "fundam.normalized"}
    elif refined:
        all_refined = set(refined)

    if "fundam" in all_refined:
        total_rows += _feat.fundam.install(
            processes=processes, recreate_tables=recreate_tables
        )

    if "fundam.normalized" in all_refined:
        total_rows += _feat.fundam.normalized.install(
            processes=processes, recreate_tables=recreate_tables
        )

    if all_ or all_refined:
        if total_rows:
            logger.info(f"{total_rows} total rows inserted for {__package__}")
        else:
            logger.warning(
                f"No rows were inserted for {__package__}. This could be an error if"
                " installations were skipped. Set the verbose flag with the"
                " `--verbose/-v` option to enable debug logging."
            )
    else:
        logger.info(
            f"Skipping {__package__} installation because no installation "
            "options are provided"
        )
    return total_rows



"""Raw features from Yahoo! Finance sources."""

import logging
import multiprocessing as mp
from datetime import datetime, timedelta

import pandas as pd
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound, NoSuchTableError
from tqdm import tqdm

from ... import backend, indices, utils
from .. import api, sql

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class Prices:
    """Get a single company's daily stock history as-is from raw Yahoo! Finance
    data.

    The module variable :data:`finagg.yfinance.feat.prices` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    """

    @classmethod
    def _install_worker(
        cls, args: tuple[str, None | str]
    ) -> tuple[None | Exception, str, pd.DataFrame]:
        """Helper for installing data from the Yahoo! Finance API using
        multiprocessing.

        """
        ticker, start = args
        try:
            df = api.get(ticker, start=start)
        except Exception as e:
            return e, ticker, pd.DataFrame()
        return None, ticker, df

    @classmethod
    def from_raw(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get a single company's daily stock history as-is from raw
        Yahoo! Finance SQL tables.

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            A dataframe containing the company's daily stock history
            across the specified period.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the raw
                SQL table.

        Examples:
            >>> finagg.yfinance.feat.prices.from_raw("AAPL").head(5)  # doctest: +SKIP
                          open    high     low   close      volume
            date
            1980-12-12  0.0997  0.1002  0.0997  0.0997  4.6903e+08
            1980-12-15  0.0950  0.0950  0.0945  0.0945  1.7588e+08
            1980-12-16  0.0880  0.0880  0.0876  0.0876  1.0573e+08
            1980-12-17  0.0897  0.0902  0.0897  0.0897  8.6442e+07
            1980-12-18  0.0924  0.0928  0.0924  0.0924  7.3450e+07

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.prices.name):
            sql.prices.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.prices.select().where(
                        sql.prices.c.ticker == ticker,
                        sql.prices.c.date >= start,
                        sql.prices.c.date <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No rows found for {ticker}.")
        return df.drop(columns=["ticker"]).set_index("date").sort_index()

    @classmethod
    def get_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique ticker symbols in the raw SQL tables that have at least
        ``lb`` rows.

        This method is convenient for accessing the tickers that have raw SQL data
        associated with them so the data associated with those tickers can be
        further refined. A common pattern is to use this method and other
        ``get_ticker_set`` methods (such as those found in :mod:`finagg.yfinance.feat`)
        to determine which tickers are missing data from other tables or features.

        Args:
            lb: Lower bound number of rows that a company must have for its ticker
                to be included in the set returned by this method.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Examples:
            >>> "AAPL" in finagg.yfinance.feat.prices.get_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.prices.name):
            sql.prices.create(engine)
        with engine.begin() as conn:
            tickers = set(
                conn.execute(
                    sa.select(sql.prices.c.ticker)
                    .where(
                        sql.prices.c.date >= start,
                        sql.prices.c.date <= end,
                    )
                    .group_by(sql.prices.c.ticker)
                    .having(sa.func.count(sql.prices.c.date) >= lb)
                )
                .scalars()
                .all()
            )
        return tickers

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from the
        API, and then writing the data to the raw prices SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.indices.api.get_ticker_set`.
            processes: Number of background processes to use when installing
                data.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's raw SQL table.

        """
        tickers = tickers or indices.api.get_ticker_set()
        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.prices.name):
            sql.prices.drop(engine, checkfirst=True)
            sql.prices.create(engine)

        total_rows = 0
        with mp.Pool(processes) as pool:
            for exc, ticker, df in tqdm(
                pool.imap_unordered(
                    cls._install_worker, [(ticker, None) for ticker in tickers]
                ),
                desc="Installing raw Yahoo! Finance stock data",
                total=len(tickers),
                position=0,
                leave=True,
            ):
                if exc:
                    logger.debug(f"Skipping {ticker}", exc_info=exc)
                    continue
                try:
                    rowcount = len(df.index)
                    if rowcount:
                        cls.to_raw(df, engine=engine)
                        total_rows += rowcount
                        logger.debug(f"{rowcount} rows inserted for {ticker}")
                    else:
                        logger.debug(f"Skipping {ticker} due to missing stock data")
                except Exception as e:
                    logger.debug(f"Skipping {ticker}", exc_info=e)
        return total_rows

    @classmethod
    def to_raw(cls, df: pd.DataFrame, /, *, engine: None | Engine = None) -> int:
        """Write the given dataframe to the raw feature table.

        Args:
            df: Dataframe to store as rows in a local SQL table
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.prices.name):
            sql.prices.create(engine)
        with engine.begin() as conn:
            conn.execute(sql.prices.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df)

    @classmethod
    def update(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
    ) -> int:
        """Update data associated with ``tickers`` by pulling data from the
        API, and then writing the data to the raw prices SQL table.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.yfinance.feat.Prices.get_ticker_set`.
            processes: Number of background processes to use when installing
                data.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the feature's raw SQL table.

        Raises:
            `NoSuchTableError`: If the table associated with this feature
                set update does not exist.

        """
        tickers = tickers or cls.get_ticker_set()
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.prices.name):
            raise NoSuchTableError(f"{sql.prices.name} table does not exist.")

        with engine.begin() as conn:
            result = conn.execute(
                sa.select(sql.prices.c.ticker, sa.func.max(sql.prices.c.date))
                .where(sql.prices.c.ticker.in_(tickers))
                .group_by(sql.prices.c.ticker)
            )
        updates = []
        for ticker, start in result:
            start_plus_one = datetime.fromisoformat(start) + timedelta(days=1)
            updates.append((ticker, start_plus_one.strftime("%Y-%m-%d")))

        total_rows = 0
        with mp.Pool(processes) as pool:
            for exc, ticker, df in tqdm(
                pool.imap_unordered(cls._install_worker, updates),
                desc="Updating raw Yahoo! Finance stock data",
                total=len(updates),
                position=0,
                leave=True,
            ):
                if exc:
                    logger.debug(f"Skipping {ticker}", exc_info=exc)
                    continue
                try:
                    rowcount = len(df.index)
                    if rowcount:
                        cls.to_raw(df, engine=engine)
                        total_rows += rowcount
                        logger.debug(f"{rowcount} rows inserted for {ticker}")
                    else:
                        logger.debug(f"Skipping {ticker} due to missing stock data")
                except Exception as e:
                    logger.debug(f"Skipping {ticker}", exc_info=e)
        return total_rows



"""Main package interface."""

from importlib.metadata import PackageNotFoundError, version

from dotenv import load_dotenv

load_dotenv()

from . import backend, bea, fred, fundam, indices, sec, testing, utils, yfinance

try:
    __version__ = version("finagg")
except PackageNotFoundError:
    pass



from typing import Generator

import pandas as pd
import pytest
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError

import finagg


@pytest.fixture
def engine() -> Generator[Engine, None, None]:
    yield from finagg.testing.sqlite_engine(
        finagg.backend.database_path, metadata=finagg.fundam.sql.metadata
    )


def test_fundam_get_candidate_ticker_set(engine: Engine) -> None:
    finagg.yfinance.feat.prices.install({"AAPL"}, engine=engine)
    finagg.yfinance.feat.daily.install({"AAPL"}, engine=engine)
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    finagg.sec.feat.quarterly.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.fundam.feat.fundam.get_candidate_ticker_set(engine=engine)


def test_fundam_get_candidate_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.fundam.feat.fundam.get_candidate_ticker_set(engine=engine)) == 0


def test_fundam_get_ticker_set(engine: Engine) -> None:
    finagg.yfinance.feat.prices.install({"AAPL"}, engine=engine)
    finagg.yfinance.feat.daily.install({"AAPL"}, engine=engine)
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    finagg.sec.feat.quarterly.install({"AAPL"}, engine=engine)
    finagg.fundam.feat.fundam.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.fundam.feat.fundam.get_ticker_set(engine=engine)


def test_fundam_get_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.fundam.feat.fundam.get_ticker_set(engine=engine)) == 0


def test_fundam_to_from_refined(engine: Engine) -> None:
    df1 = finagg.fundam.feat.fundam.from_api("AAPL")
    finagg.fundam.feat.fundam.to_refined(
        "AAPL",
        df1,
        engine=engine,
    )
    with pytest.raises(IntegrityError):
        finagg.fundam.feat.fundam.to_refined(
            "AAPL",
            df1,
            engine=engine,
        )

    df2 = finagg.fundam.feat.fundam.from_refined(
        "AAPL",
        engine=engine,
    )
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)



"""Raw features from SEC sources."""

import json
import logging
import multiprocessing as mp
from zipfile import ZipFile

import pandas as pd
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound
from tqdm import tqdm

from ... import backend, indices, utils
from .. import api, sql

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class Submissions:
    """Get a single company's metadata as-is from raw SEC data.

    The module variable :data:`finagg.sec.feat.submissions` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    """

    @classmethod
    def from_raw(
        cls,
        ticker: str,
        /,
        *,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get a single company's metadata as-is from raw SEC data.

        The metadata provided for each company varies and each company's
        metadata may be incomplete. Only their SEC CIK and SIC industry
        code are guaranteed to be provided.

        Args:
            ticker: Company ticker.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            A dataframe containing the company's metadata.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the raw
                SQL table.

        Examples:
            >>> finagg.sec.feat.submissions.from_raw("AAPL")  # doctest: +SKIP
                      cik ticker entity_type   sic sic_description ...
            0  0000320193   AAPL        None  3571            None ...

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.submissions.select().where(sql.submissions.c.ticker == ticker)
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No rows found for {ticker}.")
        return df

    @classmethod
    def get_ticker_set(
        cls,
        *,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique ticker symbols in the raw SQL submissions table.

        This method is convenient for accessing the tickers that have raw SQL data
        associated with them so the data associated with those tickers can be
        further refined. A common pattern is to use this method and other
        ``get_ticker_set`` methods (such as those found in :mod:`finagg.sec.feat`)
        to determine which tickers are missing data from other tables or features.

        Args:
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Examples:
            >>> "AAPL" in finagg.sec.feat.submissions.get_ticker_set()  # doctest: +SKIP
            True

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        with engine.begin() as conn:
            tickers = conn.execute(sa.select(sql.submissions.c.ticker)).scalars().all()
        return set(tickers)

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from the
        API, and then writing the data to the raw submissions SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.indices.api.get_ticker_set`.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        tickers = tickers or indices.api.get_ticker_set()
        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.drop(engine, checkfirst=True)
            sql.submissions.create(engine)

        total_rows = 0
        for ticker in tqdm(
            tickers,
            desc="Installing raw SEC submissions data",
            position=0,
            leave=True,
        ):
            try:
                metadata = api.submissions.get(ticker=ticker)["metadata"]
                df = pd.DataFrame(metadata, index=[0])
                rowcount = len(df.index)
                if rowcount:
                    cls.to_raw(df, engine=engine)
                    total_rows += rowcount
                    logger.debug(f"{rowcount} rows inserted for {ticker}")
                else:
                    logger.debug(f"Skipping {ticker} due to missing submissions")
            except Exception as e:
                logger.debug(f"Skipping {ticker}", exc_info=e)
        return total_rows

    @classmethod
    def install_from_zip(
        cls,
        tickers: None | set[str] = None,
        *,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install all submissions data by downloading the bulk
        submissions zip file from the API, and then writing the data to the
        raw submissions SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.indices.api.get_ticker_set`.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.drop(engine, checkfirst=True)
            sql.submissions.create(engine)

        submissions_zipfile_path = backend.root_path / "findata" / "submissions.zip"
        if recreate_tables or not submissions_zipfile_path.exists():
            zipfile = api.submissions.download_zip()
        else:
            zipfile = ZipFile(submissions_zipfile_path)

        # Filter by guaranteeing a ticker is actually present in
        # the set of tickers provided.
        tickers = tickers or indices.api.get_ticker_set()
        files = []
        for f in zipfile.namelist():
            try:
                ticker = api.get_ticker(f[3:-5])
            except KeyError:
                continue
            if ticker in tickers:
                files.append(f)

        total_rows = 0
        for f in tqdm(
            files,
            desc="Installing raw SEC submissions data",
            position=0,
            leave=True,
        ):
            try:
                cik = f[3:-5]
                ticker = api.get_ticker(cik)
                data = zipfile.read(f)
                content = json.loads(data)
                metadata = api._parse_submission_metadata(content)
                metadata["cik"] = cik
                metadata["ticker"] = ticker
                df = pd.DataFrame(metadata, index=[0])
                cls.to_raw(df, engine=engine)
                total_rows += 1
                logger.debug(f"Inserted row for {f}")
            except Exception as e:
                logger.debug(f"Skipping {f}", exc_info=e)
        return total_rows

    @classmethod
    def to_raw(cls, df: pd.DataFrame, /, *, engine: None | Engine = None) -> int:
        """Write the given dataframe to the raw feature table.

        Args:
            df: Dataframe to store as rows in a local SQL table
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        with engine.begin() as conn:
            conn.execute(sql.submissions.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df)


class Tags:
    """Get a single company concept tag as-is from raw SEC data.

    The module variable :data:`finagg.sec.feat.tags` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    """

    @classmethod
    def _install_from_zip_worker(
        cls, args: tuple[str, str]
    ) -> tuple[str, pd.DataFrame]:
        """A nasty function to make it easier for processing files within
        the company facts zip file using multiprocessing.

        Args:
            args: Tuple of the zip filename and the company facts filename
                within the zip that's being processed.

        Returns:
            A dataframe of the company's facts (empty if any error occurs during
            processing).

        """
        zip_filename, filename = args
        zipfile = ZipFile(zip_filename)
        dfs = []
        cik = filename[3:-5]
        data = zipfile.read(filename)
        content = json.loads(data)
        try:
            df = api._parse_company_facts(content)
        except:
            return filename, pd.DataFrame()
        df["cik"] = cik
        for concept in api.popular_concepts:
            try:
                tag = concept["tag"]
                taxonomy = concept["taxonomy"]
                units = concept["units"]
                df_tag = df[(df["tag"] == tag) & (df["taxonomy"] == taxonomy)]
                for form in ("10-K", "10-Q"):
                    df_unique = api.filter_original_filings(
                        df_tag, form=form, units=units
                    )
                    if len(df_unique.index):
                        dfs.append(df_unique)
            except:
                continue
        if dfs:
            return filename, pd.concat(dfs)
        return filename, pd.DataFrame()

    @classmethod
    def from_raw(
        cls,
        ticker: str,
        tag: str,
        /,
        *,
        form: str = "10-Q",
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get a single company concept tag as-is from raw SEC data.

        This is the preferred method for accessing raw SEC data without
        using the SEC API.

        Args:
            ticker: Company ticker.
            tag: Company concept tag to retreive.
            form: SEC filing form to retrieve rows for. Options include:

                - "10-Q" = quarterly filings
                - "10-K" = annual filings

            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            A dataframe containing the company concept tag values
            across the specified period.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` and ``tag``
                in the raw SQL table.

        Examples:
            >>> finagg.sec.feat.tags.from_raw("AAPL", "EarningsPerShareBasic").head(5)  # doctest: +SKIP
                                     units  value
            fy   fp filed
            2009 Q3 2009-07-22  USD/shares   4.20
            2010 Q1 2010-01-25  USD/shares   2.54
                 Q2 2010-04-21  USD/shares   4.35
                 Q3 2010-07-21  USD/shares   6.40
            2011 Q1 2011-01-19  USD/shares   3.74

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.tags.name):
            sql.tags.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sa.select(
                        sql.tags.c.fy,
                        sql.tags.c.fp,
                        sql.tags.c.filed,
                        sql.tags.c.units,
                        sql.tags.c.value,
                    )
                    .join(
                        sql.submissions,
                        (sql.submissions.c.cik == sql.tags.c.cik)
                        & (sql.submissions.c.ticker == ticker),
                    )
                    .where(
                        sql.tags.c.form == form,
                        sql.tags.c.tag == tag,
                        sql.tags.c.filed >= start,
                        sql.tags.c.filed <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No {tag} rows found for {ticker}.")
        return df.set_index(["fy", "fp", "filed"]).sort_index()

    @classmethod
    def get_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique ticker symbols in the raw SQL tables that have at least
        ``lb`` rows.

        This method is convenient for accessing the tickers that have raw SQL data
        associated with them so the data associated with those tickers can be
        further refined. A common pattern is to use this method and other
        ``get_ticker_set`` methods (such as those found in :mod:`finagg.sec.feat`)
        to determine which tickers are missing data from other tables or features.

        Args:
            lb: Lower bound number of rows that a company must have for its ticker
                to be included in the set returned by this method.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Examples:
            >>> "AAPL" in finagg.sec.feat.tags.get_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.tags.name):
            sql.tags.create(engine)
        with engine.begin() as conn:
            tickers = (
                conn.execute(
                    sa.select(sql.submissions.c.ticker)
                    .join(sql.tags, sql.tags.c.cik == sql.submissions.c.cik)
                    .where(
                        sql.tags.c.filed >= start,
                        sql.tags.c.filed <= end,
                    )
                    .group_by(sql.tags.c.cik)
                    .having(sa.func.count(sql.tags.c.filed) >= lb)
                )
                .scalars()
                .all()
            )
        return set(tickers)

    @classmethod
    def group_and_pivot_from_raw(
        cls,
        ticker: str,
        tags: list[str],
        /,
        *,
        form: str = "10-Q",
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get one or more company concept tags from raw SEC data.

        Joins all the tags into one table, pivoting the columns such that
        each tag is in its own column. Tags are forward-filled to fill
        gaps.

        Args:
            ticker: Company ticker.
            tags: Company concept tags to retreive.
            form: SEC filing form to retrieve rows for. Options include:

                - "10-Q" = quarterly filings
                - "10-K" = annual filings

            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            A dataframe containing the company concept tag values
            across the specified period.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` or any of
                the tags in ``tags`` for ``ticker`` in the raw SQL table.

        Examples:
            >>> finagg.sec.feat.tags.group_and_pivot_from_raw(
            ...     "AAPL",
            ...     ["Assets", "EarningsPerShareBasic"],
            ...     form="10-Q"
            ... ).head(5)  # doctest: +SKIP
                                      Assets  EarningsPerShareBasic
            fy   fp filed
            2009 Q3 2009-07-22  3.957200e+10                   4.20
            2010 Q1 2010-01-25  4.750100e+10                   2.54
                 Q2 2010-04-21  4.750100e+10                   4.35
                 Q3 2010-07-21  4.750100e+10                   6.40
            2011 Q1 2011-01-19  7.518300e+10                   3.74

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.tags.name):
            sql.tags.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.tags.select()
                    .join(
                        sql.submissions,
                        (sql.submissions.c.cik == sql.tags.c.cik)
                        & (sql.submissions.c.ticker == ticker),
                    )
                    .where(
                        sql.tags.c.tag.in_(tags),
                        sql.tags.c.form == form,
                        sql.tags.c.filed >= start,
                        sql.tags.c.filed <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No rows found for {ticker}.")
        df = api.group_and_pivot_filings(df, form=form)
        for tag in tags:
            if tag not in df.columns:
                raise NoResultFound(f"No {tag} rows found for {ticker}.")
        return df

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from the
        API, and then writing the data to the raw tags SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`Submissions.get_ticker_set`.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        tickers = tickers or Submissions.get_ticker_set()
        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.tags.name):
            sql.tags.drop(engine, checkfirst=True)
            sql.tags.create(engine)

        total_rows = 0
        for ticker in tqdm(
            tickers,
            desc="Installing raw SEC tags data",
            position=0,
            leave=True,
        ):
            for concept in api.popular_concepts:
                tag = concept["tag"]
                taxonomy = concept["taxonomy"]
                units = concept["units"]
                try:
                    df = api.company_concept.get(
                        tag,
                        ticker=ticker,
                        taxonomy=taxonomy,
                        units=units,
                    )
                    for form in ("10-K", "10-Q"):
                        df_unique = api.filter_original_filings(
                            df, form=form, units=units
                        )
                        rowcount = len(df_unique.index)
                        if rowcount:
                            cls.to_raw(df_unique, engine=engine)
                            total_rows += rowcount
                            logger.debug(
                                f"{rowcount} rows inserted for"
                                f" {ticker} {tag} {form} filings"
                            )
                        else:
                            logger.debug(
                                f"Skipping {ticker} due to missing {tag} {form} filings"
                            )
                except Exception as e:
                    logger.debug(f"Skipping {ticker}", exc_info=e)
        return total_rows

    @classmethod
    def install_from_zip(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install all popular tags data by downloading the bulk company
        facts zip file from the API, and then writing the data to the
        raw tags SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`Submissions.get_ticker_set`.
            processes: Number of background processes to use when installing
                data.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.tags.name):
            sql.tags.drop(engine, checkfirst=True)
            sql.tags.create(engine)

        company_facts_zipfile_path = backend.root_path / "findata" / "companyfacts.zip"
        if recreate_tables or not company_facts_zipfile_path.exists():
            zipfile = api.company_facts.download_zip()
        else:
            zipfile = ZipFile(company_facts_zipfile_path)

        # Filter by guaranteeing a ticker is actually present in
        # the set of tickers provided.
        tickers = tickers or Submissions.get_ticker_set()
        args = []
        for f in zipfile.namelist():
            try:
                ticker = api.get_ticker(f[3:-5])
            except KeyError:
                continue
            if ticker in tickers:
                args.append((zipfile.filename, f))

        total_rows = 0
        with mp.Pool(processes) as pool:
            for f, df in tqdm(
                pool.imap_unordered(cls._install_from_zip_worker, args),  # type: ignore[arg-type]
                total=len(args),
                desc="Installing raw SEC tags data",
                position=0,
                leave=True,
            ):
                try:
                    rowcount = len(df.index)
                    if rowcount:
                        cls.to_raw(df, engine=engine)
                        total_rows += rowcount
                        logger.debug(f"{rowcount} rows inserted for {f}")
                    else:
                        logger.debug(f"Skipping {f} due to missing filings")
                except Exception as e:
                    logger.debug(f"Skipping {f}", exc_info=e)
        return total_rows

    @classmethod
    def to_raw(cls, df: pd.DataFrame, /, *, engine: None | Engine = None) -> int:
        """Write the given dataframe to the raw feature table.

        Args:
            df: Dataframe to store as rows in a local SQL table
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.tags.name):
            sql.tags.create(engine)
        with engine.begin() as conn:
            conn.execute(sql.tags.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df)



import pandas as pd
import pytest
import sqlalchemy as sa

import finagg


@pytest.mark.parametrize(
    "s,expected", [("foo_bar", "FooBar"), ("FooBar", "Foobar"), ("fooBar", "Foobar")]
)
def test_CamelCase(s: str, expected: str) -> None:
    assert finagg.utils.CamelCase(s) == expected


def test_get_func_cols_from_table() -> None:
    table = sa.Table(
        "test",
        sa.MetaData(),
        sa.Column("a", sa.String, primary_key=True),
        sa.Column("LOWER(b)", sa.String, nullable=False),
        sa.Column("c", sa.String, nullable=False),
        sa.Column("UPPER(d)", sa.String, nullable=False),
    )
    assert tuple(finagg.utils.get_func_cols(table)) == ("LOWER(b)", "UPPER(d)")


def test_get_func_cols_from_df() -> None:
    df = pd.DataFrame({"a": [1], "LOWER(b)": [2], "c": [3], "UPPER(d)": [4]})
    assert tuple(finagg.utils.get_func_cols(df)) == ("LOWER(b)", "UPPER(d)")


@pytest.mark.parametrize(
    "s,expected",
    [
        ("LOG_CHANGE(high, open)", ("LOG_CHANGE", ["high", "open"])),
        ("LOG_CHANGE(open)", ("LOG_CHANGE", ["open"])),
    ],
)
def test_parse_func_call(s: str, expected: tuple[str, list[str]]) -> None:
    expected_name, expected_args = expected
    name, args = finagg.utils.parse_func_call(s)
    assert name == expected_name
    assert len(args) == len(expected_args)
    assert all([x == y for x, y in zip(args, expected_args)])


def test_resolve_col_order() -> None:
    table = sa.Table(
        "test",
        sa.MetaData(),
        sa.Column("a", sa.String, primary_key=True),
        sa.Column("b", sa.String, nullable=False),
        sa.Column("c", sa.String, nullable=False),
        sa.Column("d", sa.String, nullable=False),
    )
    df1 = pd.DataFrame(
        {"a": ["foo"], "c": ["cow"], "d": ["yes"], "b": ["bar"]}
    ).set_index("a")
    df2 = finagg.utils.resolve_col_order(table, df1)
    assert tuple(df2.columns) == ("b", "c", "d")


def test_resolve_col_order_extra_ignore() -> None:
    table = sa.Table(
        "test",
        sa.MetaData(),
        sa.Column("q", sa.String, primary_key=True),
        sa.Column("r", sa.String, nullable=False),
        sa.Column("s", sa.String, nullable=False),
        sa.Column("t", sa.String, nullable=False),
    )
    df1 = pd.DataFrame(
        {"t": ["foo"], "s": ["cow"], "r": ["yes"], "q": ["bar"]}
    ).set_index(["q", "t"])
    df2 = finagg.utils.resolve_col_order(table, df1, extra_ignore=["t"])
    assert tuple(df2.columns) == ("r", "s")


def test_safe_log_change() -> None:
    series = pd.Series([1, 2, 1])
    assert finagg.utils.safe_log_change(series).sum() == 0


def test_safe_pct_change() -> None:
    series = pd.Series([1, 2, 1])
    assert finagg.utils.safe_pct_change(series).sum() == 0.5


@pytest.mark.parametrize(
    "s,expected", [("foo_bar", "foo_bar"), ("FooBar", "foo_bar"), ("fooBar", "foo_bar")]
)
def test_snake_case(s: str, expected: str) -> None:
    assert finagg.utils.snake_case(s) == expected



"""Simple wrappers for Yahoo! Finance."""

import logging
from datetime import datetime, timedelta

import pandas as pd
import yfinance as yf

logging.getLogger("yfinance").setLevel(logging.CRITICAL)


def get(
    ticker: str,
    /,
    *,
    start: None | str = None,
    end: None | str = None,
    interval: str = "1d",
    period: str = "max",
) -> pd.DataFrame:
    """Get a ticker's stock price history.

    Does a simple transform on Yahoo! Finance's ticker API dataframe result to
    be a bit more consistent with other API implementations.

    Args:
        ticker: Company ticker to get historical price data for.
        start: Start date for stock price history. Defaults to the first
            recorded date.
        end: End date for stock price history. Defaults to the last recorded
            date.
        interval: Frequency at which stock price history is grabbed.
        period: Time period to get in the past. ``"max"`` returns the full
            stock price history and the default.

    Returns:
        Yahoo! Finance auto-adjusted stock price history with slightly
        different (more normalized) column names.

    Examples:
        >>> finagg.yfinance.api.get("AAPL").head(5)  # doctest: +SKIP
                 date    open    high     low   close     volume ticker
        0  1980-12-12  0.0997  0.1002  0.0997  0.0997  469033600   AAPL
        1  1980-12-15  0.0950  0.0950  0.0945  0.0945  175884800   AAPL
        2  1980-12-16  0.0880  0.0880  0.0876  0.0876  105728000   AAPL
        3  1980-12-17  0.0897  0.0902  0.0897  0.0897   86441600   AAPL
        4  1980-12-18  0.0924  0.0928  0.0924  0.0924   73449600   AAPL

    """
    # yfinance returns data exclusive of the end date if provided,
    # so we add an extra day to be consistent across all other
    # methods that're inclusive of the end date.
    if end is not None:
        end_plus_one = datetime.fromisoformat(end) + timedelta(days=1)
        end = end_plus_one.strftime("%Y-%m-%d")
    stock = yf.Ticker(ticker)
    df = stock.history(
        period=period,
        interval=interval,
        start=start,
        end=end,
        auto_adjust=True,
    )
    df.index = pd.to_datetime(df.index).date.astype(str)
    df = df.rename_axis("date").reset_index()
    df["ticker"] = stock.ticker
    df = df.drop(columns=["Dividends", "Stock Splits"], errors="ignore")
    df.columns = map(str.lower, df.columns)
    return df  # type: ignore[no-any-return]



import finagg


def test_category_children() -> None:
    finagg.fred.api.category.children.get(0)


def test_category_series() -> None:
    finagg.fred.api.category.series.get(10)


def test_releases_dates() -> None:
    finagg.fred.api.releases.dates.get()


def test_series() -> None:
    finagg.fred.api.series.get("GDP")


def test_series_search() -> None:
    finagg.fred.api.series.search.get("unemployment")


def test_sources() -> None:
    finagg.fred.api.sources.get()



import finagg


def test_get_cik() -> None:
    assert finagg.sec.api.get_cik("AAPL") == "0000320193"


def test_get_ticker() -> None:
    assert finagg.sec.api.get_ticker("0000320193") == "AAPL"


def test_submissions_get() -> None:
    finagg.sec.api.submissions.get(ticker="AAPL")


def test_tickers_get() -> None:
    finagg.sec.api.tickers.get()



"""Indices API (symbols in popular indices)."""

import io
import os
from abc import ABC, abstractmethod
from datetime import timedelta
from functools import cache
from typing import ClassVar

import pandas as pd
import requests
import requests_cache
from bs4 import BeautifulSoup

from .. import backend

session = requests_cache.CachedSession(
    str(backend.http_cache_path),
    expire_after=timedelta(weeks=1),
)


class API(ABC):
    """Abstract indices API."""

    #: Request API URL.
    url: ClassVar[str]

    @classmethod
    @abstractmethod
    def get(cls, *, user_agent: None | str = None) -> pd.DataFrame:
        """Main dataset API method."""

    @classmethod
    def get_ticker_list(cls, *, user_agent: None | str = None) -> list[str]:
        """List the tickers in the index."""
        df = cls.get(user_agent=user_agent)
        return df["ticker"].tolist()


class DJIA(API):
    """Get data on all companies within the DJIA.

    The module variable :data:`finagg.indices.api.djia` is an instance of
    this API implementation and is the most popular interface for querying this
    API.

    """

    url = "https://en.wikipedia.org/wiki/Dow_Jones_Industrial_Average"

    @classmethod
    def get(cls, *, user_agent: None | str = None) -> pd.DataFrame:
        """Get a dataframe containing data on the tickers in the DJIA.

        Examples:
            >>> finagg.indices.api.djia.get().head(5)  # doctest: +SKIP
                        company exchange ticker                industry       added  weight
            0                3M     NYSE    MMM            Conglomerate  1976-08-09  0.0241
            1  American Express     NYSE    AXP      Financial services  1982-08-30  0.0302
            2             Amgen   NASDAQ   AMGN       Biopharmaceutical  2020-08-31  0.0548
            3             Apple   NASDAQ   AAPL  Information technology  2015-03-19  0.0284
            4            Boeing     NYSE     BA   Aerospace and defense  1987-03-12  0.0336

        """
        response = _get(cls.url, user_agent=user_agent)
        soup = BeautifulSoup(response.text, "html.parser")
        tbl = soup.find("table", {"class": "wikitable"})
        tbl_io = io.StringIO(str(tbl))
        (df,) = pd.read_html(tbl_io)
        df = pd.DataFrame(df)

        def _percent_to_fraction(item: str) -> float:
            value, _ = item.split("%")
            return float(value) / 100

        df.drop("Notes", axis=1, inplace=True)
        df = df.rename(
            columns={
                "Company": "company",
                "Exchange": "exchange",
                "Symbol": "ticker",
                "Industry": "industry",
                "Date added": "added",
                "Index weighting": "weight",
            }
        )
        df["weight"] = df["weight"].apply(_percent_to_fraction)
        return df


class Nasdaq100(API):
    """Get data on all companies within the Nasdaq 100.

    The module variable :data:`finagg.indices.api.nasdaq100` is an instance of
    this API implementation and is the most popular interface for querying this
    API.

    """

    url = "https://en.wikipedia.org/wiki/Nasdaq-100"

    @classmethod
    def get(cls, *, user_agent: None | str = None) -> pd.DataFrame:
        """Get a dataframe containing data on the tickers in the Nasdaq 100.

        Examples:
            >>> finagg.indices.api.nasdaq100.get().head(5)  # doctest: +SKIP
                                   company ticker                industry                           sub_industry
            0          Activision Blizzard   ATVI  Communication Services         Interactive Home Entertainment
            1                   Adobe Inc.   ADBE  Information Technology                   Application Software
            2                          ADP    ADP  Information Technology  Data Processing & Outsourced Services
            3                       Airbnb   ABNB  Consumer Discretionary     Internet & Direct Marketing Retail
            4             Align Technology   ALGN             Health Care                   Health Care Supplies

        """
        response = _get(cls.url, user_agent=user_agent)
        soup = BeautifulSoup(response.text, "html.parser")
        tbl = soup.find_all("table", {"class": "wikitable"})[3]
        tbl_io = io.StringIO(str(tbl))
        (df,) = pd.read_html(tbl_io)
        df = pd.DataFrame(df)
        return df.rename(
            columns={
                "Company": "company",
                "Ticker": "ticker",
                "GICS Sector": "industry",
                "GICS Sub-Industry": "sub_industry",
            }
        )


class SP500(API):
    """Get data on all companies within the S&P 500.

    The module variable :data:`finagg.indices.api.sp500` is an instance of
    this API implementation and is the most popular interface for querying this
    API.

    """

    url = "https://en.wikipedia.org/wiki/List_of_S%26P_500_companies"

    @classmethod
    def get(cls, *, user_agent: None | str = None) -> pd.DataFrame:
        """Get a dataframe containing data on the tickers in the S&P 500.

        Examples:
            >>> finagg.indices.api.sp500.get().head(5)  # doctest: +SKIP
                ticker               company                industry ...
            0      MMM                    3M             Industrials ...
            1      AOS           A. O. Smith             Industrials ...
            2      ABT                Abbott             Health Care ...
            3     ABBV                AbbVie             Health Care ...
            4      ACN             Accenture  Information Technology ...

        """
        response = _get(cls.url, user_agent=user_agent)
        soup = BeautifulSoup(response.text, "html.parser")
        tbl = soup.find("table", {"class": "wikitable"})
        tbl_io = io.StringIO(str(tbl))
        (df,) = pd.read_html(tbl_io)
        df = pd.DataFrame(df)
        df.drop("SEC filings", axis=1, inplace=True, errors="ignore")
        return df.rename(
            columns={
                "Symbol": "ticker",
                "Security": "company",
                "GICS Sector": "industry",
                "GICS Sub-Industry": "sub_industry",
                "Headquarters Location": "headquarters",
                "Date first added": "added",
                "CIK": "cik",
                "Founded": "founded",
            }
        )


djia = DJIA()
"""The most popular way for accessing the :class:`DJIA` API
implementation.

:meta hide-value:
"""

nasdaq100 = Nasdaq100()
"""The most popular way for accessing the :class:`Nasdaq100` API
implementation.

:meta hide-value:
"""

sp500 = SP500()
"""The most popular way for accessing the :class:`SP500` API
implementation.

:meta hide-value:
"""


def _get(url: str, /, *, user_agent: None | str = None) -> requests.Response:
    """Tickers API request helper.

    Args:
        url: Complete URL to get from.
        user_agent: Required user agent header declaration to avoid errors.

    Returns:
        Successful responses.

    Raises:
        `RuntimeError`: If a user agent isn't provided or found in the environment.

    """
    user_agent = user_agent or os.environ.get(
        "INDICES_API_USER_AGENT",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/39.0.2171.95 Safari/537.36",
    )
    if not user_agent:
        raise RuntimeError(
            "No indices API user agent declaration found. "
            "Pass your user agent declaration to the API directly, or "
            "set the `INDICES_API_USER_AGENT` environment variable."
        )
    response = session.get(url, headers={"User-Agent": user_agent})
    response.raise_for_status()
    return response


@cache
def get_ticker_set(*, user_agent: None | str = None) -> set[str]:
    """Get the set of tickers from all the popular indices.

    Examples:
        >>> "AAPL" in finagg.indices.api.get_ticker_set()  # doctest: +SKIP
        True

    """
    tickers = set()
    tickers.update(djia.get_ticker_list(user_agent=user_agent))
    tickers.update(nasdaq100.get_ticker_list(user_agent=user_agent))
    tickers.update(sp500.get_ticker_list(user_agent=user_agent))
    return tickers



"""Indices SQLAlchemy interfaces."""


import sqlalchemy as sa
from sqlalchemy.engine import Engine

from .. import backend

metadata = sa.MetaData()
"""The metadata associated with all SQL tables defined in this module.

:meta hide-value:
"""

djia = sa.Table(
    "indices.raw.djia",
    metadata,
    sa.Column("company", sa.String, doc="Company name."),
    sa.Column("exchange", sa.String, doc="Exchange the company is listed on."),
    sa.Column("ticker", sa.String, primary_key=True, doc="Company ticker/symbol."),
    sa.Column("industry", sa.String, doc="The company's industry class."),
    sa.Column("added", sa.String, doc="Date the company was added to the DJIA."),
    sa.Column("weight", sa.Float, doc="Relative weight the company holds in the DJIA."),
)
"""SQL table for storing raw data as returned by
:data:`finagg.indices.api.djia` (an alias for
:class:`finagg.indices.api.DJIA`).

:meta hide-value:
"""

nasdaq100 = sa.Table(
    "indices.raw.nasdaq100",
    metadata,
    sa.Column("company", sa.String, doc="Company name."),
    sa.Column("ticker", sa.String, primary_key=True, doc="Company ticker/symbol."),
    sa.Column("industry", sa.String, doc="The company's industry class."),
    sa.Column(
        "sub_industry", sa.String, doc="The company's more specific industry class."
    ),
)
"""SQL table for storing raw data as returned by
:data:`finagg.indices.api.nasdaq100` (an alias for
:class:`finagg.indices.api.Nasdaq100`).

:meta hide-value:
"""

sp500 = sa.Table(
    "indices.raw.sp500",
    metadata,
    sa.Column("ticker", sa.String, primary_key=True, doc="Company ticker/symbol."),
    sa.Column("company", sa.String, doc="Company name."),
    sa.Column("industry", sa.String, doc="The company's industry class."),
    sa.Column(
        "sub_industry", sa.String, doc="The company's more specific industry class."
    ),
    sa.Column("headquarters", sa.String, doc="Company headquarters location."),
    sa.Column("added", sa.String, doc="Date the company was added to the S&P 500."),
    sa.Column("cik", sa.String, doc="The company's unique SEC CIK."),
    sa.Column("founded", sa.String, doc="When the company was founded."),
)
"""SQL table for storing raw data as returned by
:data:`finagg.indices.api.sp500` (an alias for
:class:`finagg.indices.api.SP500`).

:meta hide-value:
"""


def get_ticker_set(*, engine: None | Engine = None) -> set[str]:
    """Get all unique tickers in the raw SQL tables.

    Args:
        engine: Feature store database engine. Defaults to the engine
            at :data:`finagg.backend.engine`.

    Examples:
        >>> "AAPL" in finagg.indices.sql.get_ticker_set()
        True

    """
    engine = engine or backend.engine
    for table in (djia, nasdaq100, sp500):
        if not sa.inspect(engine).has_table(table.name):
            table.create(engine)
    with engine.begin() as conn:
        tickers: set[str] = set()
        for table in (djia, nasdaq100, sp500):
            tickers.update(conn.execute(sa.select(table.c.ticker)).scalars().all())
    return tickers



"""Main CLI entry points."""

import datetime
import logging
import multiprocessing as mp
import os
import pathlib
import time
from typing import Literal

import click

from . import bea, fred, fundam, indices, sec, utils, yfinance

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    ...


cli.add_command(bea._cli.entry_point, "bea")
cli.add_command(fred._cli.entry_point, "fred")
cli.add_command(fundam._cli.entry_point, "fundam")
cli.add_command(indices._cli.entry_point, "indices")
cli.add_command(sec._cli.entry_point, "sec")
cli.add_command(yfinance._cli.entry_point, "yfinance")


@cli.command(
    help=(
        "Set API keys/user agents, drop and recreate tables, "
        "and install the recommended datasets into the SQL database."
    ),
)
@click.option(
    "--skip",
    "-s",
    type=click.Choice(["bea", "fred", "indices", "sec", "yfinance", "fundam"]),
    multiple=True,
    help=(
        "Subpackage installations to skip. Useful to avoid reinstalling data "
        "that you don't want to when using the general install command."
    ),
)
@click.option(
    "--series",
    "-sid",
    multiple=True,
    help=(
        "FRED economic series whose data is attempted to be downloaded and "
        "inserted into the SQL tables. Multiple series can be specified by "
        "providing multiple `series` options, by separating series with a "
        "comma (e.g., `GDP,FEDFUNDS`), or by providing IDs in a CSV file by "
        "specifying a file path (e.g., `fred_series.txt`). The CSV file "
        "can be formatted such that there's one string per line or multiple "
        "strings per line (delimited by a comma). The strings specified "
        "by this option are combined with the strings specified by the "
        "`series-set` option."
    ),
)
@click.option(
    "--series-set",
    "-ss",
    "series_set",
    type=click.Choice(["economic"]),
    default=None,
    help=(
        "Set of FRED economic series whose data is attempted to be downloaded "
        "and inserted into the SQL tables. 'economic' indicates the recommended "
        "and most popular series (e.g., consumer price index, gross domestic "
        "product, etc.)."
    ),
)
@click.option(
    "--stock-data",
    is_flag=True,
    help=(
        "Whether to only install stock data (SEC, Yahoo! Finance, and "
        "fundamental stock data)."
    ),
)
@click.option(
    "--ticker",
    "-t",
    multiple=True,
    help=(
        "Ticker whose data is attempted to be downloaded and inserted into "
        "the SQL tables. Multiple tickers can be specified by providing "
        "multiple `ticker` options, by separating tickers with a comma (e.g., "
        "`AAPL,MSFT,NVDA`), or by providing tickers in a CSV file by "
        "specifying a file path (e.g., `bank_tickers.txt`). The CSV file "
        "can be formatted such that there's one ticker per line or multiple "
        "tickers per line (delimited by a comma). The tickers specified "
        "by this option are combined with the tickers specified by the "
        "`ticker-set` option."
    ),
)
@click.option(
    "--ticker-set",
    "-ts",
    "ticker_set",
    type=click.Choice(["indices", "sec"]),
    default=None,
    help=(
        "Set of tickers whose data is attempted to be downloaded and "
        "inserted into the SQL tables. 'indices' indicates the set "
        "of tickers from the three most popular indices (DJIA, "
        "Nasdaq 100, and S&P 500). 'sec' indicates all the tickers that "
        "have data available through the SEC API (which is approximately "
        "all publicly-traded US companies). 'indices' will effectively "
        "only attempt to download and install data for relatively "
        "popular and large market cap companies, while 'sec' will "
        "attempt to download and install data for nearly all "
        "publicly-traded US companies. Choosing 'indices' will be fast, "
        "while choosing 'sec' will be slow but will include more diverse data."
    ),
)
@click.option(
    "--from-zip",
    "-z",
    is_flag=True,
    default=False,
    help=(
        "Whether to install raw data from bulk data zip files that're compiled by the"
        " SEC nightly. Installing all SEC data with this option can take upwards of 1.5"
        " hours to complete."
    ),
)
@click.option(
    "--processes",
    "-n",
    type=int,
    default=mp.cpu_count() - 1,
    help=(
        "Number of backgruond processes to run in parallel when installing data. Note,"
        " not all tables support installations with multiprocessing."
    ),
)
@click.option(
    "--recreate-tables",
    "-r",
    is_flag=True,
    default=False,
    help=(
        "Whether to reset the tables associated with the install options by "
        "dropping and recreating them."
    ),
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Sets the log level to DEBUG to show installation errors for each ticker.",
)
@click.pass_context
def install(
    ctx: click.Context,
    skip: list[str] = [],
    series: list[str] = [],
    series_set: None | Literal["economic"] = None,
    stock_data: bool = False,
    ticker: list[str] = [],
    ticker_set: None | Literal["indices", "sec"] = None,
    from_zip: bool = False,
    processes: int = mp.cpu_count() - 1,
    recreate_tables: bool = False,
    verbose: bool = False,
) -> None:
    if "FINAGG_ROOT_PATH" not in os.environ:
        entered_root_path = input(
            "Enter the path to your finagg root directory. finagg data will "
            "be written to /path/to/root/findata/.\n\n"
            "Root path (hit ENTER to use your current working directory): "
        ).strip()
        if not entered_root_path:
            root_path = pathlib.Path.cwd()
        else:
            root_path = pathlib.Path(entered_root_path).expanduser().resolve()
        p = utils.setenv("FINAGG_ROOT_PATH", str(root_path))
        logger.info(f"FINAGG_ROOT_PATH written to {p}")
    else:
        logger.info("FINAGG_ROOT_PATH found in the environment")

    start = time.monotonic()

    all_skips = set(skip)
    if not stock_data and "bea" not in all_skips:
        ctx.invoke(bea._cli.install)

    if not stock_data and "fred" not in all_skips:
        ctx.invoke(
            fred._cli.install,
            all_=True,
            series=series,
            series_set=series_set,
            recreate_tables=recreate_tables,
            verbose=verbose,
        )

    if not stock_data and "indices" not in all_skips:
        ctx.invoke(indices._cli.install, all_=True)

    if "sec" not in all_skips:
        ctx.invoke(
            sec._cli.install,
            all_=True,
            ticker=ticker,
            ticker_set=ticker_set,
            from_zip=from_zip,
            processes=processes,
            recreate_tables=recreate_tables,
            verbose=verbose,
        )

    if "yfinance" not in all_skips:
        ctx.invoke(
            yfinance._cli.install,
            all_=True,
            ticker=ticker,
            ticker_set=ticker_set,
            processes=processes,
            recreate_tables=recreate_tables,
            verbose=verbose,
        )

    if "fundam" not in all_skips:
        ctx.invoke(
            fundam._cli.install,
            all_=True,
            processes=processes,
            recreate_tables=recreate_tables,
        )

    td = datetime.timedelta(seconds=int(time.monotonic() - start))
    logger.info(f"Installation took {td}")


def main() -> int:
    """Create and run parsers according to the given commands."""
    cli()
    return 0


if __name__ == "__main__":
    SystemExit(main())



"""`datetime` but for fiscal frames (quarter and year pairs)."""

import re
from dataclasses import dataclass
from typing import Union


def is_valid_fiscal_seq(seq: list[int], /) -> bool:
    """Determine if the sequence of fiscal quarter differences is continuous.

    A sequence that contains a jump from Q3 to Q1 is considered valid because
    companies often aren't required (and don't) report fundamentals/metrics
    for Q4.

    Args:
        seq: Sequence of fiscal quarter differences.

    Returns:
        Whether the sequence is a valid, ordered fiscal quarter sequence.

    Examples:
        This fiscal quarter difference sequence is synonymous with
        the quarter sequence [Q2, Q3, Q1, Q2, Q3].

        >>> from finagg.frame import is_valid_fiscal_seq
        >>> is_valid_fiscal_seq([1, 2, 1, 1])
        True

        This fiscal quarter difference sequence is synonymous with
        the quarter sequence [Q2, Q3, Q4, Q1, Q2].

        >>> is_valid_fiscal_seq([1, 1, 1, 1])
        True

        This fiscal quarter difference sequence is synonymous with
        the quarter sequence [Q1, Q4] which is not a valid fiscal
        sequence.

        >>> is_valid_fiscal_seq([3])
        False

    """
    valid = {(1, 1, 1), (1, 1, 2), (1, 2, 1), (2, 1, 1), (1, 1), (2, 1), (1, 2)}
    if len(seq) > 1:
        for i in range(len(seq) - 1):
            subseq = tuple(seq[i : i + 3])
            if subseq not in valid:
                return False
    else:
        if seq[0] not in {1, 2}:
            return False
    return True


@dataclass
class FiscalDelta:
    """A displacement or change from a :class:`FiscalFrame`.

    Examples:
        Get the total number of quarters from a fiscal delta.

        >>> from finagg.frame import FiscalDelta
        >>> delta = FiscalDelta(2, 2)
        >>> int(delta)
        10

    """

    #: Year delta.
    years: int = 0

    #: Quarter delta.
    quarters: int = 0

    def __int__(self) -> int:
        """Return total number of fiscal periods (quarters)."""
        return 4 * self.years + self.quarters

    def __neg__(self) -> "FiscalDelta":
        """Return the negative of the delta."""
        return FiscalDelta(-self.years, -self.quarters)

    def __post_init__(self) -> None:
        """Cast underlying data to integers."""
        self.years = int(self.years)
        self.quarters = int(self.quarters)


@dataclass
class FiscalFrame:
    """A year and quarter pair.

    Useful for comparing fiscal frames or getting fiscal frames based on fiscal
    deltas (changes in fiscal years or quarters).

    Examples:
        Getting the fiscal frame a couple years and quarters ahead.

        >>> from finagg.frame import FiscalDelta, FiscalFrame
        >>> frame = FiscalFrame(1995, 1)
        >>> frame + FiscalDelta(2, 2)
        FiscalFrame(year=1997, quarter=3)

        Adding/subtracting with integers assumes integers are quarters.

        >>> frame + 2
        FiscalFrame(year=1995, quarter=3)

        Adding/subtracting with tuples converts tuples to :class:`FiscalDelta`.

        >>> frame + (2, 2)
        FiscalFrame(year=1997, quarter=3)

        Getting quarter differences between frames and determining if the sequence
        is a valid, ordered quarterly sequence.

        >>> from finagg.frame import FiscalFrame, is_valid_fiscal_seq
        >>> df = finagg.sec.api.company_concept.get("AssetsCurrent", ticker="AAPL")
        >>> df = finagg.sec.api.filter_original_filings(df, form="10-Q", units="USD")
        >>> frames: pd.Series = df["fy"].astype(int).astype(str) + df["fp"].astype(str)
        >>> frames = frames.apply(lambda row: FiscalFrame.fromstr(row))
        >>> frames = frames.diff(periods=1).dropna().astype(int)
        >>> is_valid_fiscal_seq(frames.tolist())
        True

    """

    #: Fiscal year (e.g., 1995).
    year: int

    #: Fiscal quarter (i.e., 1, 2, 3, 4).
    quarter: int

    def __add__(self, other: object) -> "FiscalFrame":
        """Add quarters and/or years to a fiscal frame."""
        if not isinstance(other, int | FiscalDelta | tuple):
            raise TypeError(
                f"can only add {int.__name__} and {FiscalDelta.__name__} "
                f"to {self.__class__.__name__} but got "
                f"`{other.__class__.__name__}` instead."
            )

        if isinstance(other, tuple):
            years, quarters = other
            other = FiscalDelta(years, quarters)

        other = int(other)
        if other != 0:
            quarters_place = self.quarter - 1 + other
            year = self.year + (quarters_place // 4)
            quarter = (quarters_place % 4) + 1
            return FiscalFrame(year, quarter)
        return FiscalFrame(self.year, self.quarter)

    def __eq__(self, other: object) -> bool:
        """Determine if two fiscal frames are equal."""
        if not isinstance(other, FiscalFrame | tuple):
            raise TypeError(
                f"can only compare type {self.__class__.__name__} "
                f"to {self.__class__.__name__} and {tuple.__name__} but got "
                f"`{other.__class__.__name__}` instead."
            )

        if isinstance(other, tuple):
            year, quarter = other
            other = FiscalFrame(year, quarter)

        return (self.year == other.year) and (self.quarter == other.quarter)

    def __post_init__(self) -> None:
        """Argument validation and bounding."""
        quarters_place = int(self.quarter) - 1
        self.year = int(self.year) + (quarters_place // 4)
        self.quarter = (quarters_place % 4) + 1

    def __sub__(self, other: object) -> Union[FiscalDelta, "FiscalFrame"]:
        """Subtract quarters and/or years from a fiscal frame."""
        if not isinstance(other, int | FiscalDelta | FiscalFrame | tuple):
            raise TypeError(
                f"can only subtract {int.__name__}, {FiscalDelta.__name__},"
                f" {tuple.__name__}, and {self.__class__.__name__} from"
                f" {self.__class__.__name__} but got `{other.__class__.__name__}`"
                " instead."
            )

        if isinstance(other, tuple):
            years, quarters = other
            other = FiscalDelta(years, quarters)

        if isinstance(other, int | FiscalDelta):
            return self.__add__(-other)

        return FiscalDelta(self.year - other.year, self.quarter - other.quarter)

    @classmethod
    def fromstr(cls, s: str, /) -> "FiscalFrame":
        """Split a string into year-quarter parts by splitting on alphabetical characters.
        """
        year, quarter = [c for c in re.split("[a-zA-Z]", s) if c]
        return cls(int(year), int(quarter))



from typing import Generator

import pandas as pd
import pytest
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError

import finagg


@pytest.fixture
def engine() -> Generator[Engine, None, None]:
    yield from finagg.testing.sqlite_engine(
        finagg.backend.database_path, table=finagg.fred.sql.economic
    )


def test_economic_all_equal(engine: Engine) -> None:
    finagg.fred.feat.series.install(engine=engine)
    finagg.fred.feat.economic.install(engine=engine)
    df1 = finagg.fred.feat.economic.from_api().head(5)
    df2 = finagg.fred.feat.economic.from_raw(engine=engine).head(5)
    df3 = finagg.fred.feat.economic.from_refined(engine=engine).head(5)
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
    pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)


def test_economic_to_from_refined(engine: Engine) -> None:
    df1 = finagg.fred.feat.economic.from_api()
    finagg.fred.feat.economic.to_refined(
        df1,
        engine=engine,
    )
    with pytest.raises(IntegrityError):
        finagg.fred.feat.economic.to_refined(
            df1,
            engine=engine,
        )

    df2 = finagg.fred.feat.economic.from_refined(
        engine=engine,
    )
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)



"""SQLAlchemy interfaces for fundamental features."""


import sqlalchemy as sa

from .. import sec, yfinance

metadata = sa.MetaData()
"""The metadata associated with all SQL tables defined in this module.

:meta hide-value:
"""

fundam = sa.Table(
    "fundam.refined.fundam",
    metadata,
    sa.Column(
        "ticker",
        sa.String,
        sa.ForeignKey(sec.sql.submissions.c.ticker, ondelete="CASCADE"),
        sa.ForeignKey(yfinance.sql.prices.c.ticker, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique company ticker.",
    ),
    sa.Column("date", sa.String, primary_key=True, doc="Filing and stock price dates."),
    sa.Column(
        "PriceBookRatio",
        sa.Float,
        nullable=False,
        doc="Market share price over book share price.",
    ),
    sa.Column(
        "PriceEarningsRatio",
        sa.Float,
        nullable=False,
        doc="Market share price over earnings per share.",
    ),
)
"""SQL table for storing refined data as managed by
:data:`finagg.fundam.feat.fundam` (an alias for
:class:`finagg.fundam.feat.Fundamental`).

:meta hide-value:
"""

normalized_fundam = sa.Table(
    "fundam.refined.fundam.normalized",
    metadata,
    sa.Column(
        "ticker",
        sa.String,
        sa.ForeignKey(fundam.c.ticker, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique company ticker.",
    ),
    sa.Column("date", sa.String, primary_key=True, doc="Filing and stock price dates."),
    sa.Column(
        "NORM(PriceBookRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Market share price over book share price normalized against the "
            "company's industry."
        ),
    ),
    sa.Column(
        "NORM(PriceEarningsRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Market share price over earnings per share normalized against the "
            "company's industry."
        ),
    ),
)
"""SQL table for storing refined data as managed by
:attr:`finagg.fundam.feat.Fundamental.normalized` (an alias for
:class:`finagg.fundam.feat.NormalizedFundamental`).

:meta hide-value:
"""



"""Features from SEC sources."""

from ._raw import Submissions, Tags
from ._refined import (
    Annual,
    IndustryAnnual,
    IndustryQuarterly,
    NormalizedAnnual,
    NormalizedQuarterly,
    Quarterly,
)

__all__ = [
    "annual",
    "quarterly",
    "submissions",
    "tags",
    "Annual",
    "IndustryAnnual",
    "NormalizedAnnual",
    "Quarterly",
    "IndustryQuarterly",
    "NormalizedQuarterly",
    "Submissions",
    "Tags",
]

annual = Annual()
"""The most popular way for accessing :class:`finagg.sec.feat.Annual`.

:meta hide-value:
"""

quarterly = Quarterly()
"""The most popular way for accessing :class:`finagg.sec.feat.Quarterly`.

:meta hide-value:
"""

submissions = Submissions()
"""The most popular way for accessing :class:`finagg.sec.feat.Submissions`.

:meta hide-value:
"""

tags = Tags()
"""The most popular way for accessing :class:`finagg.sec.feat.Tags`.

:meta hide-value:
"""



"""Refined SEC features (features aggregated from raw tables)."""

from .annual import Annual, IndustryAnnual, NormalizedAnnual
from .quarterly import IndustryQuarterly, NormalizedQuarterly, Quarterly



from typing import Generator

import pandas as pd
import pytest
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError

import finagg


@pytest.fixture
def engine() -> Generator[Engine, None, None]:
    yield from finagg.testing.sqlite_engine(
        finagg.backend.database_path, metadata=finagg.sec.sql.metadata
    )


def test_annual_all_equal(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    finagg.sec.feat.annual.install({"AAPL"}, engine=engine)
    df1 = finagg.sec.feat.annual.from_api("AAPL").head(5)
    df2 = finagg.sec.feat.annual.from_raw("AAPL", engine=engine).head(5)
    df3 = finagg.sec.feat.annual.from_refined("AAPL", engine=engine).head(5)
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
    pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)


def test_annual_get_candidate_ticker_set(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.sec.feat.tags.get_ticker_set(engine=engine)
    assert "AAPL" in finagg.sec.feat.annual.get_candidate_ticker_set(engine=engine)


def test_annual_get_candidate_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.sec.feat.annual.get_candidate_ticker_set(engine=engine)) == 0


def test_annual_get_ticker_set(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    finagg.sec.feat.annual.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.sec.feat.annual.get_candidate_ticker_set(engine=engine)
    assert "AAPL" in finagg.sec.feat.annual.get_ticker_set(engine=engine)


def test_annual_get_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.sec.feat.annual.get_ticker_set(engine=engine)) == 0


def test_annual_to_from_refined(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    df1 = finagg.sec.feat.annual.from_api("AAPL")
    finagg.sec.feat.annual.to_refined("AAPL", df1, engine=engine)
    with pytest.raises(IntegrityError):
        finagg.sec.feat.annual.to_refined("AAPL", df1, engine=engine)

    df2 = finagg.sec.feat.annual.from_refined("AAPL", engine=engine)
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)


def test_quarterly_all_equal(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    finagg.sec.feat.quarterly.install({"AAPL"}, engine=engine)
    df1 = finagg.sec.feat.quarterly.from_api("AAPL").head(5)
    df2 = finagg.sec.feat.quarterly.from_raw("AAPL", engine=engine).head(5)
    df3 = finagg.sec.feat.quarterly.from_refined("AAPL", engine=engine).head(5)
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
    pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)


def test_quarterly_get_candidate_ticker_set(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.sec.feat.tags.get_ticker_set(engine=engine)
    assert "AAPL" in finagg.sec.feat.quarterly.get_candidate_ticker_set(engine=engine)


def test_quarterly_get_candidate_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.sec.feat.quarterly.get_candidate_ticker_set(engine=engine)) == 0


def test_quarterly_get_ticker_set(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    finagg.sec.feat.quarterly.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.sec.feat.quarterly.get_candidate_ticker_set(engine=engine)
    assert "AAPL" in finagg.sec.feat.quarterly.get_ticker_set(engine=engine)


def test_quarterly_get_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.sec.feat.quarterly.get_ticker_set(engine=engine)) == 0


def test_quarterly_to_from_refined(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    df1 = finagg.sec.feat.quarterly.from_api("AAPL")
    finagg.sec.feat.quarterly.to_refined("AAPL", df1, engine=engine)
    with pytest.raises(IntegrityError):
        finagg.sec.feat.quarterly.to_refined("AAPL", df1, engine=engine)

    df2 = finagg.sec.feat.quarterly.from_refined("AAPL", engine=engine)
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)



from typing import Generator
from unittest.mock import patch

import pytest
import requests

import finagg.ratelimit

LIMIT = 3
PERIOD = 10


class StrictLimiter(finagg.ratelimit.RateLimit):
    def eval(self, response: requests.Response) -> float | dict[str, float]:
        if hasattr(response, "from_cache") and response.from_cache:
            return 0.0
        wait = 100 if response.status_code != 200 else 0.0
        return {"limit": 1.0, "wait": wait}


@pytest.fixture
def expected_wait() -> Generator[list[int], None, None]:
    """Mock `time.perf_counter` and yield the expected test results.

    The first side effect goes to an internal dummy call.
    The response deques for each limiter before and after responses
    are filtered (assuming a limit of 3 and period of 10):

    Response appended       Responses filtered
        [0]               -> [0]
        [0, 1]            -> [0, 1]
        [0, 1, 10]        -> [1, 10]
        [1, 10, 11]       -> [10, 11]
        [10, 11, 12]      -> [10, 11, 12]  we must wait 8 seconds
        [10, 11, 12, 20]  -> [11, 12, 20]  we must wait 1 second
        [11, 12, 20, 21]  -> [12, 20, 21]  we must wait 1 second
        [12, 20, 21, 22]  -> [20, 21, 22]  we must wait 8 seconds

    """
    with patch("time.perf_counter") as perf_counter:
        perf_counter.side_effect = [0, 0, 1, 10, 11, 12, 20, 21, 22]
        yield [0, 0, 0, 0, 8, 1, 1, 8]


def test_request_limit_update(expected_wait: list[int]) -> None:
    limit = finagg.ratelimit.RequestLimit(LIMIT, PERIOD)
    response = requests.Response()
    for wait in expected_wait:
        assert limit._update(response) == wait


@patch("time.perf_counter", side_effect=[0, 0, 1, 2])
def test_request_limit_update_with_wait(_) -> None:
    limit = StrictLimiter(LIMIT, PERIOD)
    ok_response = requests.Response()
    ok_response.status_code = 200
    bad_response = requests.Response()
    bad_response.status_code = 404
    assert limit._update(ok_response) == 0.0
    assert limit._update(bad_response) == 100.0
    assert limit._update(ok_response) == 99.0


def test_error_limit_update(expected_wait: list[int]) -> None:
    limit = finagg.ratelimit.ErrorLimit(LIMIT, PERIOD)
    response = requests.Response()
    response.status_code = 404
    for wait in expected_wait:
        assert limit._update(response) == wait


def test_size_limit_update(expected_wait: list[int]) -> None:
    limit = finagg.ratelimit.SizeLimit(LIMIT, PERIOD)
    response = requests.Response()
    response._content = b"0"
    for wait in expected_wait:
        assert limit._update(response) == wait



"""Features from Yahoo! Finance sources."""

from ._raw import Prices
from ._refined import Daily

__all__ = ["daily", "prices", "Daily", "Prices"]

daily = Daily()
"""The most popular way for accessing :class:`finagg.yfinance.feat.Daily`.

:meta hide-value:
"""

prices = Prices()
"""The most popular way for accessing :class:`finagg.yfinance.feat.Prices`.

:meta hide-value:
"""



"""Generic utils used by subpackages."""

import csv
import logging
import multiprocessing as mp
import os
import pathlib
import re
from datetime import datetime
from pathlib import Path
from typing import Protocol

import numpy as np
import pandas as pd
import sqlalchemy as sa
from dotenv import set_key
from tqdm import tqdm


def CamelCase(s: str, /) -> str:
    """Transform a string to CamelCase.

    Credit:
        https://stackoverflow.com/a/1176023

    Args:
        s: Any string.

    Returns:
        A string in CamelCase format.

    Examples:
        >>> finagg.utils.CamelCase("snakes_are_dope") == "SnakesAreDope"
        True
        >>> finagg.utils.CamelCase("bar") == "Bar"
        True

    """
    return "".join(word.title() for word in s.split("_"))


def expand_csv(values: str | list[str], /) -> set[str]:
    """Expand the given list of strings into a set of strings, where each value
    in the list of strings could be:

        1. Comma-separated values
        2. A path that points to a CSV file containing values
        3. A regular ol' string

    Args:
        values: List of strings denoting comma-separated values,
            or CSV files containing comma-separated values.

    Returns:
        A set of all strings found within the given list.

    Examples:
        >>> ts = finagg.utils.expand_csv(["AAPL,MSFT"])
        >>> "AAPL" in ts
        True

    """
    if isinstance(values, str):
        values = [values]

    out = set()
    for vstring in values:
        vlist = vstring.split(",")
        for v in vlist:
            csv_path = Path(v)
            if csv_path.exists():
                with open(csv_path, "r") as f:
                    reader = csv.reader(f)
                    for row in reader:
                        out.update(row)
            else:
                out.add(v)
    return out


def get_func_cols(table: sa.Table | pd.DataFrame, /) -> list[str]:
    """Return the column names in ``table`` that have the format
    ``FUNC(arg0, arg1, ...)``.

    Args:
        table: SQLAlchemy table or dataframe.

    Returns:
        List of functional-style column names in ``table``. Returns
        an empty list if none are found.

    Raises:
        `TypeError`: If the given object is not a SQLAlchemy
            table or dataframe.

    """
    match table:
        case sa.Table():
            cols = table.columns.keys()
        case pd.DataFrame():
            cols = table.columns.to_list()
        case _:
            raise TypeError(f"{table.__class__.__name__} is not supported")
    return [col for col in cols if parse_func_call(col)]


def parse_func_call(s: str, /) -> None | tuple[str, list[str]]:
    """Parse a function's name and its arguments' names from a string of format
    ``FUNC(arg0, arg1, ...)``.

    Args:
        s: Any string of format ``FUNC(arg0, arg1, ...)``.

    Returns:
        A tuple containing the parsed function's name and its arguments' names.
        Returns ``None`` if the string doesn't match the expected format.

    Examples:
        >>> finagg.utils.parse_func_call("LOG_CHANGE(high, open)")
        ('LOG_CHANGE', ['high', 'open'])

    """
    match = re.match(r"(\w+)\((.*)\)", s)
    if not match:
        return None
    name, args = match.groups()
    return name, args.replace(" ", "").split(",")


def resolve_col_order(
    table: sa.Table, df: pd.DataFrame, /, *, extra_ignore: None | list[str] = None
) -> pd.DataFrame:
    """Reorder the columns in ``df`` to match the order of the columns in
    ``table``.

    Args:
        table: SQLAlchemy table that defines the column order. Primary keys
            are ignored from the column order as they're assumed to be used
            as part of the index in ``df``.
        df: Dataframe to reorder.
        extra_ignore: Extra columns to ignore in the reordering. Sometimes
            columns aren't used as primary keys but are used as part of the
            index in the dataframe. Those columns should be provided in this
            option.

    Returns:
        Dataframe with columns ordered according to the column order in
        ``table``.

    """
    if extra_ignore is None:
        extra_ignore = []
    ignore_keys = {col.key for col in table.primary_key}
    ignore_keys.update(extra_ignore)
    column_order = [key for key in table.columns.keys() if key not in ignore_keys]
    return df[column_order]


def resolve_func_cols(
    table: sa.Table, df: pd.DataFrame, /, *, drop: bool = False, inplace: bool = False
) -> pd.DataFrame:
    """Inspect ``table`` and apply functions to columns that exist in ``table``
    and ``df`` according to columns named like ``FUNC(col0, col1, ...)``
    within ``table`` such that new columns in ``df`` are the result of the
    applied functions and have names matching the function call signatures.

    Args:
        table: SQLAchemy table that defines a superset of columns that
            should exist in ``df``.
        df: Dataframe that contains a subset of columns within ``table``
            that will be updated with columns defined by ``table`` that
            have names like ``FUNC(col0, col1, ...)``.
        drop: Whether to drop all other columns on the returned dataframe
            except for the columns in ``table``.
        inplace: Whether to perform operations in-place and use ``df``
            as the output dataframe.

    Returns:
        A new dataframe with columns from ``df`` and columns according to
        columns named within ``table`` like ``FUNC(col0, col1, ...)`` where
        columns ``col0`` and ``col1`` exist in ``df``.

    Raises:
        `ValueError`: If the function parsed from the column name has no
            supported and corresponding function.

    """
    out = df if inplace else df.copy(deep=True)
    primary_keys = {col.key for col in table.primary_key}
    other_keys = [key for key in table.columns.keys() if key not in primary_keys]
    func_keys = set()
    for key in other_keys:
        if func_call := parse_func_call(key):
            func_keys.add(key)
            name, args = func_call
            cols = map(out.get, args)
            match name:
                case "LOG_CHANGE":
                    out[key] = safe_log_change(*cols)  # type: ignore[arg-type]
                case "PCT_CHANGE":
                    out[key] = safe_pct_change(*cols)  # type: ignore[arg-type]
                case _:
                    raise ValueError(f"{key} is not supported")
    if inplace and drop:
        out.drop(columns=list(set(out.columns) - set(other_keys)), inplace=True)
    elif drop:
        out = out.drop(columns=list(set(out.columns) - set(other_keys)))
    return out


def safe_log_change(series: pd.Series, other: None | pd.Series = None) -> pd.Series:  # type: ignore[type-arg]
    """Safely compute log change between two columns.

    Replaces ``Inf`` values with ``NaN`` and forward-fills.
    This function is meant to be used with ``pd.Series.apply``.

    Args:
        series: Series of values.
        other: Reference series to compute change against. Defaults to
            ``series`` shifted forward one index.

    Returns:
        A series representing percent changes of ``col``.

    """
    if other is None:
        other = series.shift(1)

    out = series.apply(np.log) - other.apply(np.log)
    return out.replace([-np.inf, np.inf], np.nan).ffill()


def safe_pct_change(series: pd.Series, other: None | pd.Series = None) -> pd.Series:  # type: ignore[type-arg]
    """Safely compute percent change between two columns.

    Replaces ``Inf`` values with ``NaN`` and forward-fills.
    This function is meant to be used with ``pd.Series.apply``.

    Args:
        series: Series of values.
        other: Reference series to compute change against. Defaults to
            ``series`` shifted forward one index.

    Returns:
        A series representing percent changes of ``col``.

    """
    if other is None:
        other = series.shift(1)

    out = (series - other) / other
    return out.replace([-np.inf, np.inf], np.nan).ffill()


def setenv(name: str, value: str, /, *, exist_ok: bool = False) -> pathlib.Path:
    """Set the value of the environment variable ``name`` to ``value``.

    The environment variable is permanently set in the environment
    and in the current process.

    Args:
        name: Environment variable name.
        value: Environment variable value.
        exist_ok: Whether it's okay if an environment variable of the
            same name already exists. If ``True``, it will be overwritten.

    Returns:
        Path to the file the environment variable was written to.

    Raises:
        `RuntimeError`: If ``exist_ok`` is ``False`` and an environment variable
            of the same name already exists.

    """
    if not exist_ok and name in os.environ:
        raise RuntimeError(
            f"The env variable `{name}` already exists. "
            "Set `exist_ok` to `True` to overwrite it."
        )

    os.environ[name] = value
    dotenv = pathlib.Path.cwd() / ".env"
    set_key(str(dotenv), name, value)
    return dotenv


def snake_case(s: str, /) -> str:
    """Transform a string to snake_case.

    Credit:
        https://stackoverflow.com/a/1176023

    Args:
        s: Any string.

    Returns:
        A string in snake_case format.

    Examples:
        >>> finagg.utils.snake_case("CamelsAreCool") == "camels_are_cool"
        True
        >>> finagg.utils.snake_case("Foo") == "foo"
        True

    """
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    s = re.sub("__([A-Z])", r"_\1", s)
    s = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s)
    return s.lower()


today = datetime.today().strftime("%Y-%m-%d")
"""Today's date. Used by a number of submodules as the default end date
when getting data from APIs or SQL tables.

:meta hide-value:
"""


class _ReadFn(Protocol):
    @classmethod
    def __call__(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | sa.Engine = None,
    ) -> pd.DataFrame:
        ...


class _WriteFn(Protocol):
    @classmethod
    def __call__(
        cls, ticker: str, df: pd.DataFrame, /, *, engine: None | sa.Engine = None
    ) -> int:
        ...


class _InstallWorkerFn(Protocol):
    @classmethod
    def __call__(cls, ticker: str, engine: None | sa.Engine = None) -> pd.DataFrame:
        ...


class _InstallWorker:
    engine: None | sa.Engine = None

    fn: None | _InstallWorkerFn = None

    @classmethod
    def init(cls, url: str | sa.URL, fn: _InstallWorkerFn) -> None:
        cls.engine = sa.create_engine(url)
        cls.fn = fn

    @classmethod
    def call(cls, ticker: str) -> tuple[None | Exception, str, pd.DataFrame]:
        assert cls.engine is not None
        assert cls.fn is not None
        try:
            df = cls.fn(ticker, engine=cls.engine)
        except Exception as e:
            return e, ticker, pd.DataFrame()
        return None, ticker, df


def _install(
    read_fn: _ReadFn,
    write_fn: _WriteFn,
    logger: logging.Logger,
    tickers: list[str],
    engine: sa.Engine,
    /,
    *,
    desc: None | str = None,
    processes: int = mp.cpu_count() - 1,
) -> int:
    """Helper for feature installation methods.

    Useful for reducing code duplication, but this helper
    may not exist in the future.

    """
    total_rows = 0
    with (
        mp.Pool(
            processes,
            initializer=_InstallWorker.init,
            initargs=(engine.url, read_fn),
        ) as pool,
        tqdm(
            total=len(tickers),
            desc=desc,
            position=0,
            leave=True,
        ) as pb,
    ):
        for group in (
            tickers[i : i + processes] for i in range(0, len(tickers), processes)
        ):
            results = []
            for exc, ticker, df in pool.imap_unordered(_InstallWorker.call, group):
                if exc:
                    logger.debug(f"Skipping {ticker}", exc_info=exc)
                    pb.update()
                else:
                    results.append((ticker, df))
            for ticker, df in results:
                try:
                    rowcount = len(df.index)
                    if rowcount:
                        write_fn(ticker, df, engine=engine)
                        total_rows += rowcount
                        logger.debug(f"{rowcount} rows inserted for {ticker}")
                    else:
                        logger.debug(f"Skipping {ticker} due to missing data")
                except Exception as e:
                    logger.debug(f"Skipping {ticker}", exc_info=e)
                pb.update()
    return total_rows



"""An implementation of the Bureau of Economic Analysis (BEA) API.

The BEA API provides methods for retrieving a subset of economic
statistical data as published by the BEA along with metadata that
describes that economic statistical data.

A BEA API key is required to use this API. You can register for
a BEA API key at the `BEA API signup page`_. You can pass your
BEA API key directly to the implemented API getters, or you
can set the ``BEA_API_KEY`` environment variable to have the
BEA API key be passed to the implemented API getters for you.

Alternatively, running ``finagg bea install`` (or the broader
``finagg install``) will  prompt you where to aquire a BEA API
key and will automatically store it in an ``.env`` file in
your current working directory. The environment variables set
in that ``.env`` file will be loaded into your shell upon
using ``finagg`` (whether that be through the Python interface
or through the CLI tools).

See the official `BEA API user guide`_ for more info on the BEA API.

.. _`BEA API signup page`: https://apps.bea.gov/api/signup/
.. _`BEA API user guide`: https://apps.bea.gov/api/_pdf/bea_web_service_api_user_guide.pdf

.. note::
    This was the first API implementation in this project, but has since
    lost priority in favor of the FRED API as the FRED API provides
    data that's found through the BEA API in addition to a plethora
    of other data. This BEA API implementation is still maintained
    and supported, but other features such as data installation are not
    and will never be supported.

"""

import json
import logging
import os
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Any, ClassVar, Literal, Sequence

import pandas as pd
import requests
import requests_cache

from .. import backend, ratelimit

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

if backend.disable_http_cache:
    session = requests.Session()
else:
    session = requests_cache.CachedSession(
        str(backend.http_cache_path),
        ignored_parameters=["ResultFormat"],
        expire_after=timedelta(days=1),
    )

_YEAR = int | str


class API(ABC):
    """Interface for BEA Dataset APIs."""

    #: Request API URL.
    name: ClassVar[str]

    @classmethod
    @abstractmethod
    def get(cls, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """Main dataset API method."""

    @classmethod
    def get_parameter_list(cls, /, *, api_key: None | str = None) -> pd.DataFrame:
        """Return the list of parameters associated with the dataset API."""
        return _get_parameter_list(cls.name, api_key=api_key)

    @classmethod
    def get_parameter_values(
        cls, param: str, /, *, api_key: None | str = None
    ) -> pd.DataFrame:
        """Return all possible parameter values associated with the dataset API."""
        return _get_parameter_values(cls.name, param, api_key=api_key)


class FixedAssets(API):
    """US fixed assets (assets for long-term use)."""

    name = "FixedAssets"

    @classmethod
    def get(
        cls,
        table_id: str | Sequence[str] = "ALL",
        year: _YEAR | Sequence[_YEAR] = "ALL",
        *,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get US fixed assets by asset and year.

        Args:
            table_id: IDs associated with assets of concern.
                Use :meth:`~finagg.bea.api.API.get_parameter_values` to see
                possible values.
            year: Years to return.

        Returns:
            Dataframe with normalized column names and true dtypes.

        """
        if table_id == "ALL":
            table_id = cls.get_parameter_values("TableName")["TableName"].to_list()
        elif isinstance(table_id, str):
            table_id = [table_id]

        results = []
        for tid in table_id:
            params = {
                "Method": "GetData",
                "DatasetName": cls.name,
                "TableName": tid,
                "Year": year,
            }
            results_data = _get(params, api_key=api_key)
            data = results_data["Data"]
            df = (
                pd.DataFrame(data)
                .drop("NoteRef", axis=1)
                .rename(
                    columns={
                        "TableName": "table_id",
                        "SeriesCode": "series_code",
                        "LineNumber": "line",
                        "LineDescription": "line_description",
                        "TimePeriod": "year",
                        "METRIC_NAME": "metric",
                        "CL_UNIT": "units",
                        "UNIT_MULT": "e",
                        "DataValue": "value",
                    }
                )
                .astype(
                    {
                        "table_id": "category",
                        "series_code": "category",
                        "line": "int16",
                        "line_description": "object",
                        "year": "int16",
                        "metric": "category",
                        "units": "category",
                        "e": "int16",
                    }
                )
            )
            df["value"] = df["value"].str.replace(",", "").astype("float32")
            results.append(df)
        return pd.concat(results)


class GDPByIndustry(API):
    """GDP (a single summary statistic) for each industry.

    The module variable :data:`finagg.bea.api.gdp_by_industry` is an instance
    of this API implementation and is the most popular interface for querying
    this API.

    Data provided by this API is considered coarse/high-level.
    See :class:`InputOutput` for more granular/low-level industry data.

    Examples:
        List the GDP by industry API parameters.

        >>> finagg.bea.api.gdp_by_industry.get_parameter_list()  # doctest: +SKIP
          ParameterName ParameterDataType                               ParameterDescription ... AllValue
        0     Frequency            string                            A - Annual, Q-Quarterly ...      ALL
        1      Industry            string       List of industries to retrieve (ALL for All) ...      ALL
        2       TableID           integer  The unique GDP by Industry table identifier (A... ...      ALL
        3          Year           integer  List of year(s) of data to retrieve (ALL for All) ...      ALL

        List possible GDP by industry tables we can query.

        >>> finagg.bea.api.gdp_by_industry.get_parameter_values("TableID").head(5)  # doctest: +SKIP
          Key                                               Desc
        0   1                    Value Added by Industry (A) (Q)
        1   5  Value added by Industry as a Percentage of Gro...
        2   6          Components of Value Added by Industry (A)
        3   7  Components of Value Added by Industry as a Per...
        4   8  Chain-Type Quantity Indexes for Value Added by...

    """

    name = "GdpByIndustry"

    @classmethod
    def get(
        cls,
        table_id: str | Sequence[str] = "ALL",
        freq: Literal["A", "Q", "A,Q"] = "Q",
        year: _YEAR | Sequence[_YEAR] = "ALL",
        industry: str | Sequence[str] = "ALL",
        *,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get GDP by industry.

        Args:
            table_id: IDs associated with GDP value type. Use
                :meth:`~finagg.bea.api.API.get_parameter_values` to see
                possible values. `"ALL"` indicates retrieve all GDP value
                measurement type tables.
            freq: Data frequency to return. `"Q"` for quarterly, `"A"` for
                annually, and `"A,Q"` for both annually and quarterly.
            year: Years to return. `"ALL"` indicates retrieve data for all
                available years.
            industry: IDs associated with industries. Use
                :meth:`~finagg.bea.api.API.get_parameter_values` to see
                possible values.

        Returns:
            Dataframe with GDP by industry, separated by year and/or quarter.

        Examples:
            Get the GDP value added by an industry for a specific year.

            >>> finagg.bea.api.gdp_by_industry.get(table_id=1, freq="A", year=2020).head(5)  # doctest: +SKIP
               table_id freq  year quarter industry                         industry_description  value
            0         1    A  2020    2020       11  Agriculture, forestry, fishing, and hunting  162.2
            1         1    A  2020    2020    111CA                                        Farms  120.7
            2         1    A  2020    2020    113FF    Forestry, fishing, and related activities   41.5
            3         1    A  2020    2020       21                                       Mining  201.1
            4         1    A  2020    2020      211                       Oil and gas extraction  110.9

        """
        params = {
            "Method": "GetData",
            "DatasetName": cls.name,
            "TableID": table_id,
            "Frequency": freq,
            "Year": year,
            "Industry": industry,
        }
        (results_data,) = _get(params, api_key=api_key)
        data = results_data["Data"]  # type: ignore
        df = pd.DataFrame(data)

        if freq != "A":

            def _roman_to_int(item: str) -> int:
                _map = {"I": 1, "II": 2, "III": 3, "IV": 4}
                return _map[item]

            df["Quarter"] = df["Quarter"].apply(_roman_to_int)

        df.drop("NoteRef", axis=1, inplace=True)
        return df.rename(
            columns={
                "TableID": "table_id",
                "Frequency": "freq",
                "Year": "year",
                "Quarter": "quarter",
                "Industry": "industry",
                "IndustrYDescription": "industry_description",
                "DataValue": "value",
            }
        ).astype(
            {
                "table_id": "int16",
                "freq": "category",
                "year": "int16",
                "quarter": "category",
                "industry": "category",
                "industry_description": "object",
                "value": "float32",
            }
        )


class InputOutput(API):
    """Specific input-output statistics for each industry.

    Data provided by this API is considered granular/low-level.
    See :class:`GDPByIndustry` for more coarse/high-level industry data.

    Data is provided for different "rows" and "columns" where:
        - a row is an industry and
        - a column is a statistic associated with that industry

    Columns are divided by column codes. Each industry of similar
    type has the same set of column codes that provide input-output
    statistics for that industry.

    """

    name = "InputOutput"

    @classmethod
    def get(
        cls,
        table_id: str | Sequence[str] = "ALL",
        year: _YEAR | Sequence[_YEAR] = "ALL",
        *,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get input-output statistics by industry.

        Args:
            table_id: IDs associated with input-output stats. Use
                :meth:`~finagg.bea.api.API.get_parameter_values` to see
                possible values. `"ALL"` indicates retrieve all tables for
                all types of input-output statistics by industry.
            year: Years to return. `"ALL"` indicates retrieve data for all
                available years.

        Returns:
            Dataframe organized by table row and column codes.

        """
        params = {
            "Method": "GetData",
            "DatasetName": cls.name,
            "TableID": table_id,
            "Year": year,
        }
        (results_data,) = _get(params, api_key=api_key)
        data = results_data["Data"]  # type: ignore
        return (
            pd.DataFrame(data)
            .drop("NoteRef", axis=1)
            .rename(
                columns={
                    "TableID": "table_id",
                    "Year": "year",
                    "RowCode": "row_code",
                    "RowDescr": "row_description",
                    "RowType": "row_type",
                    "ColCode": "col_code",
                    "ColDescr": "col_description",
                    "ColType": "col_type",
                    "DataValue": "value",
                }
            )
            .astype(
                {
                    "table_id": "int16",
                    "year": "int16",
                    "row_code": "category",
                    "row_description": "object",
                    "row_type": "category",
                    "col_code": "category",
                    "col_description": "object",
                    "col_type": "category",
                    "value": "float32",
                }
            )
        )


class NIPA(API):
    """National income and product accounts.

    Details high-level US economic details in several
    metrics.

    """

    name = "NIPA"

    @classmethod
    def get(
        cls,
        table_id: str | Sequence[str] = "ALL",
        freq: Literal["A", "Q", "A,Q"] = "Q",
        year: _YEAR | Sequence[_YEAR] = "ALL",
        *,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get US income and product accounts by metric.

        Args:
            table_id: IDs associated with metric of concern.
                Use :meth:`~finagg.bea.api.API.get_parameter_values` to see
                possible values.
            freq: Data frequency to return. `"Q"` for quarterly, `"A"` for annually.
            year: Years to return.

        Returns:
            Dataframe with normalized column names and true dtypes.

        """
        if table_id == "ALL":
            table_id = cls.get_parameter_values("TableName")["TableName"].to_list()
        elif isinstance(table_id, str):
            table_id = [table_id]

        results = []
        for tid in table_id:
            params = {
                "Method": "GetData",
                "DatasetName": cls.name,
                "TableName": tid,
                "Year": year,
                "Frequency": freq,
            }
            results_data = _get(params, api_key=api_key)
            data = results_data["Data"]
            df = pd.DataFrame(data)
            df[["Year", "Quarter"]] = df["TimePeriod"].str.split("Q", n=1, expand=True)
            df["Quarter"] = df["Quarter"].astype(int)
            df.drop(["TimePeriod", "NoteRef"], axis=1, inplace=True)
            df = df.rename(
                columns={
                    "TableName": "table_id",
                    "SeriesCode": "series_code",
                    "LineNumber": "line",
                    "LineDescription": "line_description",
                    "Year": "year",
                    "Quarter": "quarter",
                    "METRIC_NAME": "metric",
                    "CL_UNIT": "units",
                    "UNIT_MULT": "e",
                    "DataValue": "value",
                }
            ).astype(
                {
                    "table_id": "category",
                    "series_code": "category",
                    "line": "int16",
                    "line_description": "object",
                    "year": "int16",
                    "quarter": "int16",
                    "metric": "category",
                    "units": "category",
                    "e": "int16",
                }
            )
            df["value"] = df["value"].str.replace(",", "").astype("float32")
            results.append(df)
        return pd.concat(results)


fixed_assets = FixedAssets()
"""The most popular way for accessing the :class:`finagg.bea.api.FixedAssets`
API implementation.

:meta hide-value:
"""

gdp_by_industry = GDPByIndustry()
"""The most popular way for accessing the :class:`finagg.bea.api.GDPByIndustry`
API implementation.

:meta hide-value:
"""

input_output = InputOutput()
"""The most popular way for accessing the :class:`finagg.bea.api.InputOutput`
API implementation.

:meta hide-value:
"""

nipa = NIPA()
"""The most popular way for accessing the :class:`finagg.bea.api.NIPA`
API implementation.

:meta hide-value:
"""

#: The BEA API endpoint URL. All API requests are made to this URL.
url = "https://apps.bea.gov/api/data"


def _api_error_as_response(error: dict[str, str]) -> requests.Response:
    """Convert an API error to a :class:`requests.Response` object."""
    response = requests.Response()
    response.status_code = int(error.pop("APIErrorCode"))
    response._content = json.dumps(error).encode("utf-8")
    return response


def _get(
    params: dict[str, Any],
    /,
    *,
    api_key: None | str = None,
) -> dict[str, list[dict[str, Any]]]:
    """Main get method used by dataset APIs.

    Handles API key validation, formatting, and parameters common
    amongst several API methods.

    Args:
        params: Params specific to the API method.

    Returns:
        Results from the API request (typically a list of records).

    Raises:
        `RuntimeError`: If no BEA API key is passed or found.
        `BEAAPIError`: If a BEA API error occurs.

    """
    api_key = api_key or os.environ.get("BEA_API_KEY", None)
    if not api_key:
        raise RuntimeError(
            "No BEA API key found. "
            "Pass the API key to the API directly, or "
            "set the `BEA_API_KEY` environment variable."
        )

    params.update({"UserID": api_key, "ResultFormat": "JSON"})
    response = _guarded_get(url, params)
    response.raise_for_status()
    content = response.json()["BEAAPI"]
    if "Error" in content:
        error = _api_error_as_response(content["Error"])
        raise BEAAPIError(response.request, error, error.content)
    if "Error" in content["Results"]:
        error = _api_error_as_response(content["Results"]["Error"])
        raise BEAAPIError(response.request, error, error.content)
    return content["Results"]  # type: ignore


def _get_parameter_list(dataset: str, /, *, api_key: None | str = None) -> pd.DataFrame:
    """Get a dataset's list of parameters.

    Args:
        dataset: Dataset API to inspect. See :meth:`get_dataset_list` for a
            list of datasets.

    Returns:
        Dataframe listing the dataset's parameters.

    """
    params = {
        "Method": "GetParameterList",
        "DatasetName": dataset,
    }
    results = _get(params, api_key=api_key)["Parameter"]
    return pd.DataFrame(results)


def _get_parameter_values(
    dataset: str, param: str, /, *, api_key: None | str = None
) -> pd.DataFrame:
    """Get potential values for a dataset's parameter.

    Args:
        dataset: Dataset API to inspect. See :meth:`get_dataset_list` for a
            list of datasets.
        param: Dataset API's parameter to inspect.

    Returns:
        Dataframe describing the dataset's parameter values.

    """
    params = {
        "Method": "GetParameterValues",
        "DatasetName": dataset,
        "ParameterName": param,
    }
    results = _get(params, api_key=api_key)["ParamValue"]
    return pd.DataFrame(results)


@ratelimit.guard(
    [
        ratelimit.RequestLimit(90, timedelta(minutes=1)),
        ratelimit.ErrorLimit(20, timedelta(minutes=1)),
        ratelimit.SizeLimit(90e6, timedelta(minutes=1)),
    ],
)
def _guarded_get(url: str, params: dict[str, Any], /) -> requests.Response:
    """Guarded version of `session.get`."""
    return session.get(url, params=params)


def get_dataset_list(*, api_key: None | str = None) -> pd.DataFrame:
    """Return a list of datasets provided by the BEA API.

    Returns:
        A dataframe describing the datasets available through the BEA API.

    Examples:
        >>> finagg.bea.api.get_dataset_list()  # doctest: +SKIP
                        DatasetName                    DatasetDescription
        0                      NIPA                  Standard NIPA tables
        1        NIUnderlyingDetail  Standard NI underlying detail tables
        2                       MNE             Multinational Enterprises
        3               FixedAssets          Standard Fixed Assets tables
        4                       ITA   International Transactions Accounts
        5                       IIP     International Investment Position
        6               InputOutput                     Input-Output Data
        7             IntlServTrade          International Services Trade
        8             GDPbyIndustry                       GDP by Industry
        9                  Regional                    Regional data sets
        10  UnderlyingGDPbyIndustry            Underlying GDP by Industry
        11       APIDatasetMetaData     Metadata about other API datasets

    """
    params = {
        "Method": "GetDatasetList",
    }
    results = _get(params, api_key=api_key)["Dataset"]
    return pd.DataFrame(results)


class BEAAPIError(requests.RequestException):
    """An error raised in response to BEA API request errors."""
